# Copyright (C) 2015-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import itertools
import os
import re

from collections import defaultdict
from typing import Any, Dict, List, Set, Iterator, Optional, Tuple

from swh.model import hashutil
from swh.model.identifiers import CONTENT, DIRECTORY, RELEASE, REVISION, SNAPSHOT
from swh.storage.algos import diff, revisions_walker
from swh.vault.exc import NotFoundExc as VaultNotFoundExc
from swh.web import config
from swh.web.common import converters
from swh.web.common import query
from swh.web.common.exc import BadInputExc, NotFoundExc
from swh.web.common.origin_visits import get_origin_visit
from swh.web.common.typing import OriginInfo, OriginVisitInfo


search = config.search()
storage = config.storage()
vault = config.vault()
idx_storage = config.indexer_storage()


MAX_LIMIT = 50  # Top limit the users can ask for


def _first_element(lst):
    """Returns the first element in the provided list or None
    if it is empty or None"""
    return next(iter(lst or []), None)


def lookup_multiple_hashes(hashes):
    """Lookup the passed hashes in a single DB connection, using batch
    processing.

    Args:
        An array of {filename: X, sha1: Y}, string X, hex sha1 string Y.
    Returns:
        The same array with elements updated with elem['found'] = true if
        the hash is present in storage, elem['found'] = false if not.

    """
    hashlist = [hashutil.hash_to_bytes(elem["sha1"]) for elem in hashes]
    content_missing = storage.content_missing_per_sha1(hashlist)
    missing = [hashutil.hash_to_hex(x) for x in content_missing]
    for x in hashes:
        x.update({"found": True})
    for h in hashes:
        if h["sha1"] in missing:
            h["found"] = False
    return hashes


def lookup_expression(expression, last_sha1, per_page):
    """Lookup expression in raw content.

    Args:
        expression (str): An expression to lookup through raw indexed
        content
        last_sha1 (str): Last sha1 seen
        per_page (int): Number of results per page

    Yields:
        ctags whose content match the expression

    """

    limit = min(per_page, MAX_LIMIT)
    ctags = idx_storage.content_ctags_search(
        expression, last_sha1=last_sha1, limit=limit
    )
    for ctag in ctags:
        ctag = converters.from_swh(ctag, hashess={"id"})
        ctag["sha1"] = ctag["id"]
        ctag.pop("id")
        yield ctag


def lookup_hash(q):
    """Checks if the storage contains a given content checksum

    Args: query string of the form <hash_algo:hash>

    Returns: Dict with key found containing the hash info if the
    hash is present, None if not.

    """
    algo, hash = query.parse_hash(q)
    found = _first_element(storage.content_find({algo: hash}))
    return {"found": converters.from_content(found), "algo": algo}


def search_hash(q):
    """Checks if the storage contains a given content checksum

    Args: query string of the form <hash_algo:hash>

    Returns: Dict with key found to True or False, according to
        whether the checksum is present or not

    """
    algo, hash = query.parse_hash(q)
    found = _first_element(storage.content_find({algo: hash}))
    return {"found": found is not None}


def _lookup_content_sha1(q):
    """Given a possible input, query for the content's sha1.

    Args:
        q: query string of the form <hash_algo:hash>

    Returns:
        binary sha1 if found or None

    """
    algo, hash = query.parse_hash(q)
    if algo != "sha1":
        hashes = _first_element(storage.content_find({algo: hash}))
        if not hashes:
            return None
        return hashes["sha1"]
    return hash


def lookup_content_ctags(q):
    """Return ctags information from a specified content.

    Args:
        q: query string of the form <hash_algo:hash>

    Yields:
        ctags information (dict) list if the content is found.

    """
    sha1 = _lookup_content_sha1(q)

    if not sha1:
        return None

    ctags = list(idx_storage.content_ctags_get([sha1]))
    if not ctags:
        return None

    for ctag in ctags:
        yield converters.from_swh(ctag, hashess={"id"})


def lookup_content_filetype(q):
    """Return filetype information from a specified content.

    Args:
        q: query string of the form <hash_algo:hash>

    Yields:
        filetype information (dict) list if the content is found.

    """
    sha1 = _lookup_content_sha1(q)
    if not sha1:
        return None
    filetype = _first_element(list(idx_storage.content_mimetype_get([sha1])))
    if not filetype:
        return None
    return converters.from_filetype(filetype)


def lookup_content_language(q):
    """Return language information from a specified content.

    Args:
        q: query string of the form <hash_algo:hash>

    Yields:
        language information (dict) list if the content is found.

    """
    sha1 = _lookup_content_sha1(q)
    if not sha1:
        return None
    lang = _first_element(list(idx_storage.content_language_get([sha1])))
    if not lang:
        return None
    return converters.from_swh(lang, hashess={"id"})


def lookup_content_license(q):
    """Return license information from a specified content.

    Args:
        q: query string of the form <hash_algo:hash>

    Yields:
        license information (dict) list if the content is found.

    """
    sha1 = _lookup_content_sha1(q)
    if not sha1:
        return None
    lic = _first_element(idx_storage.content_fossology_license_get([sha1]))

    if not lic:
        return None
    return converters.from_swh({"id": sha1, "facts": lic[sha1]}, hashess={"id"})


def lookup_origin(origin: OriginInfo) -> OriginInfo:
    """Return information about the origin matching dict origin.

    Args:
        origin: origin's dict with 'url' key

    Returns:
        origin information as dict.

    """
    origins = [origin]
    if origin["url"]:
        # handle case when user provided an origin url with a trailing
        # slash while the url in storage does not have it (e.g. GitHub)
        if origin["url"].endswith("/"):
            origins.append({"url": origin["url"][:-1]})
        # handle case when user provided an origin url without a trailing
        # slash while the url in storage have it (e.g. Debian source package)
        else:
            origins.append({"url": f"{origin['url']}/"})
    # Check all possible origin urls
    for orig in origins:
        origin_info = storage.origin_get(orig)
        if origin_info:
            break
    if not origin_info:
        msg = "Origin with url %s not found!" % origin["url"]
        raise NotFoundExc(msg)
    return converters.from_origin(origin_info)


def lookup_origins(
    origin_from: int = 1, origin_count: int = 100
) -> Iterator[OriginInfo]:
    """Get list of archived software origins in a paginated way.

    Origins are sorted by id before returning them

    Args:
        origin_from (int): The minimum id of the origins to return
        origin_count (int): The maximum number of origins to return

    Yields:
        origins information as dicts
    """
    origins = storage.origin_get_range(origin_from, origin_count)
    return map(converters.from_origin, origins)


def search_origin(
    url_pattern: str, limit: int = 50, with_visit: bool = False, page_token: Any = None
) -> Tuple[List[OriginInfo], Any]:
    """Search for origins whose urls contain a provided string pattern
    or match a provided regular expression.

    Args:
        url_pattern: the string pattern to search for in origin urls
        limit: the maximum number of found origins to return
        page_token: opaque string used to get the next results of a search

    Returns:
        list of origin information as dict.

    """
    if search:
        results = search.origin_search(
            url_pattern=url_pattern,
            count=limit,
            page_token=page_token,
            with_visit=with_visit,
        )
        origins = list(map(converters.from_origin, results["results"]))
        return (origins, results["next_page_token"])
    else:
        # Fallback to swh-storage if swh-search is not configured
        offset = int(page_token) if page_token else 0
        regexp = True
        search_words = [re.escape(word) for word in url_pattern.split()]
        if len(search_words) >= 7:
            url_pattern = ".*".join(search_words)
        else:
            pattern_parts = []
            for permut in itertools.permutations(search_words):
                pattern_parts.append(".*".join(permut))
            url_pattern = "|".join(pattern_parts)

        origins_raw = storage.origin_search(
            url_pattern, offset, limit, regexp, with_visit
        )
        origins = list(map(converters.from_origin, origins_raw))
        if len(origins) >= limit:
            page_token = str(offset + len(origins))
        else:
            page_token = None
        return (origins, page_token)


def search_origin_metadata(fulltext, limit=50):
    """Search for origins whose metadata match a provided string pattern.

    Args:
        fulltext: the string pattern to search for in origin metadata
        offset: number of found origins to skip before returning results
        limit: the maximum number of found origins to return

    Returns:
        list of origin metadata as dict.

    """
    matches = idx_storage.origin_intrinsic_metadata_search_fulltext(
        conjunction=[fulltext], limit=limit
    )
    results = []

    for match in matches:
        match["from_revision"] = hashutil.hash_to_hex(match["from_revision"])

        origin = storage.origin_get({"url": match["id"]})
        del match["id"]

        result = converters.from_origin(origin)
        if result:
            result["metadata"] = match
            results.append(result)

    return results


def lookup_origin_intrinsic_metadata(origin_dict):
    """Return intrinsic metadata for origin whose origin matches given
    origin.

    Args:
        origin_dict: origin's dict with keys ('type' AND 'url')

    Returns:
        origin metadata.

    """
    origin_info = storage.origin_get(origin_dict)
    if not origin_info:
        msg = "Origin with url %s not found!" % origin_dict["url"]
        raise NotFoundExc(msg)

    origins = [origin_info["url"]]
    match = _first_element(idx_storage.origin_intrinsic_metadata_get(origins))
    result = {}
    if match:
        result = match["metadata"]
    return result


def _to_sha1_bin(sha1_hex):
    _, sha1_git_bin = query.parse_hash_with_algorithms_or_throws(
        sha1_hex, ["sha1"], "Only sha1_git is supported."  # HACK: sha1_git really
    )
    return sha1_git_bin


def _check_directory_exists(sha1_git, sha1_git_bin):
    if len(list(storage.directory_missing([sha1_git_bin]))):
        raise NotFoundExc("Directory with sha1_git %s not found" % sha1_git)


def lookup_directory(sha1_git):
    """Return information about the directory with id sha1_git.

    Args:
        sha1_git as string

    Returns:
        directory information as dict.

    """
    empty_dir_sha1 = "4b825dc642cb6eb9a060e54bf8d69288fbee4904"

    if sha1_git == empty_dir_sha1:
        return []

    sha1_git_bin = _to_sha1_bin(sha1_git)

    _check_directory_exists(sha1_git, sha1_git_bin)

    directory_entries = storage.directory_ls(sha1_git_bin)
    return map(converters.from_directory_entry, directory_entries)


def lookup_directory_with_path(sha1_git, path_string):
    """Return directory information for entry with path path_string w.r.t.
    root directory pointed by directory_sha1_git

    Args:
        - directory_sha1_git: sha1_git corresponding to the directory
        to which we append paths to (hopefully) find the entry
        - the relative path to the entry starting from the directory pointed by
        directory_sha1_git

    Raises:
        NotFoundExc if the directory entry is not found
    """
    sha1_git_bin = _to_sha1_bin(sha1_git)

    _check_directory_exists(sha1_git, sha1_git_bin)

    paths = path_string.strip(os.path.sep).split(os.path.sep)
    queried_dir = storage.directory_entry_get_by_path(
        sha1_git_bin, list(map(lambda p: p.encode("utf-8"), paths))
    )

    if not queried_dir:
        raise NotFoundExc(
            ("Directory entry with path %s from %s not found") % (path_string, sha1_git)
        )

    return converters.from_directory_entry(queried_dir)


def lookup_release(release_sha1_git):
    """Return information about the release with sha1 release_sha1_git.

    Args:
        release_sha1_git: The release's sha1 as hexadecimal

    Returns:
        Release information as dict.

    Raises:
        ValueError if the identifier provided is not of sha1 nature.

    """
    sha1_git_bin = _to_sha1_bin(release_sha1_git)
    release = _first_element(storage.release_get([sha1_git_bin]))
    if not release:
        raise NotFoundExc("Release with sha1_git %s not found." % release_sha1_git)
    return converters.from_release(release)


def lookup_release_multiple(sha1_git_list):
    """Return information about the revisions identified with
    their sha1_git identifiers.

    Args:
        sha1_git_list: A list of revision sha1_git identifiers

    Returns:
        Release information as dict.

    Raises:
        ValueError if the identifier provided is not of sha1 nature.

    """
    sha1_bin_list = (_to_sha1_bin(sha1_git) for sha1_git in sha1_git_list)
    releases = storage.release_get(sha1_bin_list) or []
    return (converters.from_release(r) for r in releases)


def lookup_revision(rev_sha1_git):
    """Return information about the revision with sha1 revision_sha1_git.

    Args:
        revision_sha1_git: The revision's sha1 as hexadecimal

    Returns:
        Revision information as dict.

    Raises:
        ValueError if the identifier provided is not of sha1 nature.
        NotFoundExc if there is no revision with the provided sha1_git.

    """
    sha1_git_bin = _to_sha1_bin(rev_sha1_git)
    revision = _first_element(storage.revision_get([sha1_git_bin]))
    if not revision:
        raise NotFoundExc("Revision with sha1_git %s not found." % rev_sha1_git)
    return converters.from_revision(revision)


def lookup_revision_multiple(sha1_git_list):
    """Return information about the revisions identified with
    their sha1_git identifiers.

    Args:
        sha1_git_list: A list of revision sha1_git identifiers

    Returns:
        Iterator of revisions information as dict.

    Raises:
        ValueError if the identifier provided is not of sha1 nature.

    """
    sha1_bin_list = (_to_sha1_bin(sha1_git) for sha1_git in sha1_git_list)
    revisions = storage.revision_get(sha1_bin_list) or []
    return (converters.from_revision(r) for r in revisions)


def lookup_revision_message(rev_sha1_git):
    """Return the raw message of the revision with sha1 revision_sha1_git.

    Args:
        revision_sha1_git: The revision's sha1 as hexadecimal

    Returns:
        Decoded revision message as dict {'message': <the_message>}

    Raises:
        ValueError if the identifier provided is not of sha1 nature.
        NotFoundExc if the revision is not found, or if it has no message

    """
    sha1_git_bin = _to_sha1_bin(rev_sha1_git)

    revision = _first_element(storage.revision_get([sha1_git_bin]))
    if not revision:
        raise NotFoundExc("Revision with sha1_git %s not found." % rev_sha1_git)
    if "message" not in revision:
        raise NotFoundExc("No message for revision with sha1_git %s." % rev_sha1_git)
    res = {"message": revision["message"]}
    return res


def _lookup_revision_id_by(origin, branch_name, timestamp):
    def _get_snapshot_branch(snapshot, branch_name):
        snapshot = lookup_snapshot(
            visit["snapshot"], branches_from=branch_name, branches_count=10
        )
        branch = None
        if branch_name in snapshot["branches"]:
            branch = snapshot["branches"][branch_name]
        return branch

    if isinstance(origin, int):
        origin = {"id": origin}
    elif isinstance(origin, str):
        origin = {"url": origin}
    else:
        raise TypeError('"origin" must be an int or a string.')

    visit = get_origin_visit(origin, visit_ts=timestamp)
    branch = _get_snapshot_branch(visit["snapshot"], branch_name)
    rev_id = None
    if branch and branch["target_type"] == "revision":
        rev_id = branch["target"]
    elif branch and branch["target_type"] == "alias":
        branch = _get_snapshot_branch(visit["snapshot"], branch["target"])
        if branch and branch["target_type"] == "revision":
            rev_id = branch["target"]

    if not rev_id:
        raise NotFoundExc(
            "Revision for origin %s and branch %s not found."
            % (origin.get("url"), branch_name)
        )

    return rev_id


def lookup_revision_by(origin, branch_name="HEAD", timestamp=None):
    """Lookup revision by origin, snapshot branch name and visit timestamp.

    If branch_name is not provided, lookup using 'HEAD' as default.
    If timestamp is not provided, use the most recent.

    Args:
        origin (Union[int,str]): origin of the revision
        branch_name (str): snapshot branch name
        timestamp (str/int): origin visit time frame

    Returns:
        dict: The revision matching the criterions

    Raises:
        NotFoundExc if no revision corresponds to the criterion

    """
    rev_id = _lookup_revision_id_by(origin, branch_name, timestamp)
    return lookup_revision(rev_id)


def lookup_revision_log(rev_sha1_git, limit):
    """Lookup revision log by revision id.

    Args:
        rev_sha1_git (str): The revision's sha1 as hexadecimal
        limit (int): the maximum number of revisions returned

    Returns:
        list: Revision log as list of revision dicts

    Raises:
        ValueError: if the identifier provided is not of sha1 nature.
        swh.web.common.exc.NotFoundExc: if there is no revision with the
            provided sha1_git.

    """
    lookup_revision(rev_sha1_git)
    sha1_git_bin = _to_sha1_bin(rev_sha1_git)
    revision_entries = storage.revision_log([sha1_git_bin], limit)
    return map(converters.from_revision, revision_entries)


def lookup_revision_log_by(origin, branch_name, timestamp, limit):
    """Lookup revision by origin, snapshot branch name and visit timestamp.

    Args:
        origin (Union[int,str]): origin of the revision
        branch_name (str): snapshot branch
        timestamp (str/int): origin visit time frame
        limit (int): the maximum number of revisions returned

    Returns:
        list: Revision log as list of revision dicts

    Raises:
        swh.web.common.exc.NotFoundExc: if no revision corresponds to the
            criterion

    """
    rev_id = _lookup_revision_id_by(origin, branch_name, timestamp)
    return lookup_revision_log(rev_id, limit)


def lookup_revision_with_context_by(
    origin, branch_name, timestamp, sha1_git, limit=100
):
    """Return information about revision sha1_git, limited to the
    sub-graph of all transitive parents of sha1_git_root.
    sha1_git_root being resolved through the lookup of a revision by origin,
    branch_name and ts.

    In other words, sha1_git is an ancestor of sha1_git_root.

    Args:
        - origin: origin of the revision.
        - branch_name: revision's branch.
        - timestamp: revision's time frame.
        - sha1_git: one of sha1_git_root's ancestors.
        - limit: limit the lookup to 100 revisions back.

    Returns:
        Pair of (root_revision, revision).
        Information on sha1_git if it is an ancestor of sha1_git_root
        including children leading to sha1_git_root

    Raises:
        - BadInputExc in case of unknown algo_hash or bad hash.
        - NotFoundExc if either revision is not found or if sha1_git is not an
        ancestor of sha1_git_root.

    """
    rev_root_id = _lookup_revision_id_by(origin, branch_name, timestamp)

    rev_root_id_bin = hashutil.hash_to_bytes(rev_root_id)

    rev_root = _first_element(storage.revision_get([rev_root_id_bin]))

    return (
        converters.from_revision(rev_root),
        lookup_revision_with_context(rev_root, sha1_git, limit),
    )


def lookup_revision_with_context(sha1_git_root, sha1_git, limit=100):
    """Return information about revision sha1_git, limited to the
    sub-graph of all transitive parents of sha1_git_root.

    In other words, sha1_git is an ancestor of sha1_git_root.

    Args:
        sha1_git_root: latest revision. The type is either a sha1 (as an hex
        string) or a non converted dict.
        sha1_git: one of sha1_git_root's ancestors
        limit: limit the lookup to 100 revisions back

    Returns:
        Information on sha1_git if it is an ancestor of sha1_git_root
        including children leading to sha1_git_root

    Raises:
        BadInputExc in case of unknown algo_hash or bad hash
        NotFoundExc if either revision is not found or if sha1_git is not an
        ancestor of sha1_git_root

    """
    sha1_git_bin = _to_sha1_bin(sha1_git)

    revision = _first_element(storage.revision_get([sha1_git_bin]))
    if not revision:
        raise NotFoundExc("Revision %s not found" % sha1_git)

    if isinstance(sha1_git_root, str):
        sha1_git_root_bin = _to_sha1_bin(sha1_git_root)

        revision_root = _first_element(storage.revision_get([sha1_git_root_bin]))
        if not revision_root:
            raise NotFoundExc("Revision root %s not found" % sha1_git_root)
    else:
        sha1_git_root_bin = sha1_git_root["id"]

    revision_log = storage.revision_log([sha1_git_root_bin], limit)

    parents = {}
    children = defaultdict(list)

    for rev in revision_log:
        rev_id = rev["id"]
        parents[rev_id] = []
        for parent_id in rev["parents"]:
            parents[rev_id].append(parent_id)
            children[parent_id].append(rev_id)

    if revision["id"] not in parents:
        raise NotFoundExc(
            "Revision %s is not an ancestor of %s" % (sha1_git, sha1_git_root)
        )

    revision["children"] = children[revision["id"]]

    return converters.from_revision(revision)


def lookup_directory_with_revision(sha1_git, dir_path=None, with_data=False):
    """Return information on directory pointed by revision with sha1_git.
    If dir_path is not provided, display top level directory.
    Otherwise, display the directory pointed by dir_path (if it exists).

    Args:
        sha1_git: revision's hash.
        dir_path: optional directory pointed to by that revision.
        with_data: boolean that indicates to retrieve the raw data if the path
        resolves to a content. Default to False (for the api)

    Returns:
        Information on the directory pointed to by that revision.

    Raises:
        BadInputExc in case of unknown algo_hash or bad hash.
        NotFoundExc either if the revision is not found or the path referenced
        does not exist.
        NotImplementedError in case of dir_path exists but do not reference a
        type 'dir' or 'file'.

    """
    sha1_git_bin = _to_sha1_bin(sha1_git)
    revision = _first_element(storage.revision_get([sha1_git_bin]))
    if not revision:
        raise NotFoundExc("Revision %s not found" % sha1_git)
    dir_sha1_git_bin = revision["directory"]
    if dir_path:
        paths = dir_path.strip(os.path.sep).split(os.path.sep)
        entity = storage.directory_entry_get_by_path(
            dir_sha1_git_bin, list(map(lambda p: p.encode("utf-8"), paths))
        )
        if not entity:
            raise NotFoundExc(
                "Directory or File '%s' pointed to by revision %s not found"
                % (dir_path, sha1_git)
            )
    else:
        entity = {"type": "dir", "target": dir_sha1_git_bin}
    if entity["type"] == "dir":
        directory_entries = storage.directory_ls(entity["target"]) or []
        return {
            "type": "dir",
            "path": "." if not dir_path else dir_path,
            "revision": sha1_git,
            "content": list(map(converters.from_directory_entry, directory_entries)),
        }
    elif entity["type"] == "file":  # content
        content = _first_element(storage.content_find({"sha1_git": entity["target"]}))
        if not content:
            raise NotFoundExc("Content not found for revision %s" % sha1_git)
        if with_data:
            c = _first_element(storage.content_get([content["sha1"]]))
            content["data"] = c["data"]
        return {
            "type": "file",
            "path": "." if not dir_path else dir_path,
            "revision": sha1_git,
            "content": converters.from_content(content),
        }
    elif entity["type"] == "rev":  # revision
        revision = next(storage.revision_get([entity["target"]]))
        return {
            "type": "rev",
            "path": "." if not dir_path else dir_path,
            "revision": sha1_git,
            "content": converters.from_revision(revision),
        }
    else:
        raise NotImplementedError("Entity of type %s not implemented." % entity["type"])


def lookup_content(q):
    """Lookup the content designed by q.

    Args:
        q: The release's sha1 as hexadecimal

    Raises:
        NotFoundExc if the requested content is not found

    """
    algo, hash = query.parse_hash(q)
    c = _first_element(storage.content_find({algo: hash}))
    if not c:
        raise NotFoundExc(
            "Content with %s checksum equals to %s not found!"
            % (algo, hashutil.hash_to_hex(hash))
        )
    return converters.from_content(c)


def lookup_content_raw(q):
    """Lookup the content defined by q.

    Args:
        q: query string of the form <hash_algo:hash>

    Returns:
        dict with 'sha1' and 'data' keys.
        data representing its raw data decoded.

    Raises:
        NotFoundExc if the requested content is not found or
        if the content bytes are not available in the storage

    """
    c = lookup_content(q)
    content_sha1_bytes = hashutil.hash_to_bytes(c["checksums"]["sha1"])
    content = _first_element(storage.content_get([content_sha1_bytes]))
    if not content:
        algo, hash = query.parse_hash(q)
        raise NotFoundExc(
            "Bytes of content with %s checksum equals to %s "
            "are not available!" % (algo, hashutil.hash_to_hex(hash))
        )
    return converters.from_content(content)


def stat_counters():
    """Return the stat counters for Software Heritage

    Returns:
        A dict mapping textual labels to integer values.
    """
    return storage.stat_counters()


def _lookup_origin_visits(
    origin_url: str, last_visit: Optional[int] = None, limit: int = 10
) -> Iterator[Dict[str, Any]]:
    """Yields the origin origins' visits.

    Args:
        origin_url (str): origin to list visits for
        last_visit (int): last visit to lookup from
        limit (int): Number of elements max to display

    Yields:
       Dictionaries of origin_visit for that origin

    """
    limit = min(limit, MAX_LIMIT)
    for visit in storage.origin_visit_get(
        origin_url, last_visit=last_visit, limit=limit
    ):
        visit["origin"] = origin_url
        yield visit


def lookup_origin_visits(
    origin: str, last_visit: Optional[int] = None, per_page: int = 10
) -> Iterator[OriginVisitInfo]:
    """Yields the origin origins' visits.

    Args:
        origin: origin to list visits for

    Yields:
       Dictionaries of origin_visit for that origin

    """
    visits = _lookup_origin_visits(origin, last_visit=last_visit, limit=per_page)
    for visit in visits:
        yield converters.from_origin_visit(visit)


def lookup_origin_visit_latest(
    origin_url: str, require_snapshot: bool
) -> OriginVisitInfo:
    """Return the origin's latest visit

    Args:
        origin_url (str): origin to list visits for
        require_snapshot (bool): filter out origins without a snapshot

    Returns:
       dict: The origin_visit concerned

    """
    visit = storage.origin_visit_get_latest(
        origin_url, require_snapshot=require_snapshot
    )
    return converters.from_origin_visit(visit)


def lookup_origin_visit(origin_url: str, visit_id: int) -> OriginVisitInfo:
    """Return information about visit visit_id with origin origin.

    Args:
        origin (str): origin concerned by the visit
        visit_id: the visit identifier to lookup

    Yields:
       The dict origin_visit concerned

    """
    visit = storage.origin_visit_get_by(origin_url, visit_id)
    if not visit:
        raise NotFoundExc(
            "Origin %s or its visit " "with id %s not found!" % (origin_url, visit_id)
        )
    visit["origin"] = origin_url
    return converters.from_origin_visit(visit)


def lookup_snapshot_sizes(snapshot_id):
    """Count the number of branches in the snapshot with the given id

    Args:
        snapshot_id (str): sha1 identifier of the snapshot

    Returns:
        dict: A dict whose keys are the target types of branches and
        values their corresponding amount
    """
    snapshot_id_bin = _to_sha1_bin(snapshot_id)
    snapshot_sizes = storage.snapshot_count_branches(snapshot_id_bin)
    if "revision" not in snapshot_sizes:
        snapshot_sizes["revision"] = 0
    if "release" not in snapshot_sizes:
        snapshot_sizes["release"] = 0
    # adjust revision / release count for display if aliases are defined
    if "alias" in snapshot_sizes:
        aliases = lookup_snapshot(
            snapshot_id, branches_count=snapshot_sizes["alias"], target_types=["alias"]
        )
        for alias in aliases["branches"].values():
            if lookup_snapshot(
                snapshot_id,
                branches_from=alias["target"],
                branches_count=1,
                target_types=["revision"],
            ):
                snapshot_sizes["revision"] += 1
            else:
                snapshot_sizes["release"] += 1
        del snapshot_sizes["alias"]
    return snapshot_sizes


def lookup_snapshot(
    snapshot_id, branches_from="", branches_count=1000, target_types=None
):
    """Return information about a snapshot, aka the list of named
    branches found during a specific visit of an origin.

    Args:
        snapshot_id (str): sha1 identifier of the snapshot
        branches_from (str): optional parameter used to skip branches
            whose name is lesser than it before returning them
        branches_count (int): optional parameter used to restrain
            the amount of returned branches
        target_types (list): optional parameter used to filter the
            target types of branch to return (possible values that can be
            contained in that list are `'content', 'directory',
            'revision', 'release', 'snapshot', 'alias'`)

    Returns:
        A dict filled with the snapshot content.
    """
    snapshot_id_bin = _to_sha1_bin(snapshot_id)
    snapshot = storage.snapshot_get_branches(
        snapshot_id_bin, branches_from.encode(), branches_count, target_types
    )
    if not snapshot:
        raise NotFoundExc("Snapshot with id %s not found!" % snapshot_id)
    return converters.from_snapshot(snapshot)


def lookup_latest_origin_snapshot(origin, allowed_statuses=None):
    """Return information about the latest snapshot of an origin.

    .. warning:: At most 1000 branches contained in the snapshot
        will be returned for performance reasons.

    Args:
        origin: URL or integer identifier of the origin
        allowed_statuses: list of visit statuses considered
            to find the latest snapshot for the visit. For instance,
            ``allowed_statuses=['full']`` will only consider visits that
            have successfully run to completion.

    Returns:
        A dict filled with the snapshot content.
    """
    snapshot = storage.snapshot_get_latest(origin, allowed_statuses)
    return converters.from_snapshot(snapshot)


def lookup_snapshot_branch_name_from_tip_revision(
    snapshot_id: str, revision_id: str
) -> Optional[str]:
    """Check if a revision corresponds to the tip of a snapshot branch

    Args:
        snapshot_id: hexadecimal representation of a snapshot id
        revision_id: hexadecimal representation of a revision id

    Returns:
        The name of the first found branch or None otherwise
    """
    per_page = 10000
    branches_from = ""
    snapshot: Dict[str, Any] = {"branches": {}}
    branches = []
    while not branches_from or len(snapshot["branches"]) == per_page + 1:
        snapshot = lookup_snapshot(
            snapshot_id,
            target_types=[REVISION],
            branches_from=branches_from,
            branches_count=per_page + 1,
        )

        branches += [
            {"name": k, "target": v["target"]} for k, v in snapshot["branches"].items()
        ]
        branches_from = branches[-1]["name"]

    for branch in branches:
        if branch["target"] == revision_id:
            return branch["name"]
    return None


def lookup_revision_through(revision, limit=100):
    """Retrieve a revision from the criterion stored in revision dictionary.

    Args:
        revision: Dictionary of criterion to lookup the revision with.
        Here are the supported combination of possible values:
        - origin_url, branch_name, ts, sha1_git
        - origin_url, branch_name, ts
        - sha1_git_root, sha1_git
        - sha1_git

    Returns:
        None if the revision is not found or the actual revision.

    """
    if (
        "origin_url" in revision
        and "branch_name" in revision
        and "ts" in revision
        and "sha1_git" in revision
    ):
        return lookup_revision_with_context_by(
            revision["origin_url"],
            revision["branch_name"],
            revision["ts"],
            revision["sha1_git"],
            limit,
        )
    if "origin_url" in revision and "branch_name" in revision and "ts" in revision:
        return lookup_revision_by(
            revision["origin_url"], revision["branch_name"], revision["ts"]
        )
    if "sha1_git_root" in revision and "sha1_git" in revision:
        return lookup_revision_with_context(
            revision["sha1_git_root"], revision["sha1_git"], limit
        )
    if "sha1_git" in revision:
        return lookup_revision(revision["sha1_git"])

    # this should not happen
    raise NotImplementedError("Should not happen!")


def lookup_directory_through_revision(revision, path=None, limit=100, with_data=False):
    """Retrieve the directory information from the revision.

    Args:
        revision: dictionary of criterion representing a revision to lookup
        path: directory's path to lookup.
        limit: optional query parameter to limit the revisions log (default to
            100). For now, note that this limit could impede the transitivity
            conclusion about sha1_git not being an ancestor of.
        with_data: indicate to retrieve the content's raw data if path resolves
            to a content.

    Returns:
        The directory pointing to by the revision criterions at path.

    """
    rev = lookup_revision_through(revision, limit)

    if not rev:
        raise NotFoundExc("Revision with criterion %s not found!" % revision)
    return (rev["id"], lookup_directory_with_revision(rev["id"], path, with_data))


def _vault_request(vault_fn, *args, **kwargs):
    try:
        return vault_fn(*args, **kwargs)
    except VaultNotFoundExc:
        return None


def vault_cook(obj_type, obj_id, email=None):
    """Cook a vault bundle.
    """
    return _vault_request(vault.cook, obj_type, obj_id, email=email)


def vault_fetch(obj_type, obj_id):
    """Fetch a vault bundle.
    """
    return _vault_request(vault.fetch, obj_type, obj_id)


def vault_progress(obj_type, obj_id):
    """Get the current progress of a vault bundle.
    """
    return _vault_request(vault.progress, obj_type, obj_id)


def diff_revision(rev_id):
    """Get the list of file changes (insertion / deletion / modification /
    renaming) for a particular revision.
    """
    rev_sha1_git_bin = _to_sha1_bin(rev_id)

    changes = diff.diff_revision(storage, rev_sha1_git_bin, track_renaming=True)

    for change in changes:
        change["from"] = converters.from_directory_entry(change["from"])
        change["to"] = converters.from_directory_entry(change["to"])
        if change["from_path"]:
            change["from_path"] = change["from_path"].decode("utf-8")
        if change["to_path"]:
            change["to_path"] = change["to_path"].decode("utf-8")

    return changes


class _RevisionsWalkerProxy(object):
    """
    Proxy class wrapping a revisions walker iterator from
    swh-storage and performing needed conversions.
    """

    def __init__(self, rev_walker_type, rev_start, *args, **kwargs):
        rev_start_bin = hashutil.hash_to_bytes(rev_start)
        self.revisions_walker = revisions_walker.get_revisions_walker(
            rev_walker_type, storage, rev_start_bin, *args, **kwargs
        )

    def export_state(self):
        return self.revisions_walker.export_state()

    def __next__(self):
        return converters.from_revision(next(self.revisions_walker))

    def __iter__(self):
        return self


def get_revisions_walker(rev_walker_type, rev_start, *args, **kwargs):
    """
    Utility function to instantiate a revisions walker of a given type,
    see :mod:`swh.storage.algos.revisions_walker`.

    Args:
        rev_walker_type (str): the type of revisions walker to return,
            possible values are: ``committer_date``, ``dfs``, ``dfs_post``,
            ``bfs`` and ``path``
        rev_start (str): hexadecimal representation of a revision identifier
        args (list): position arguments to pass to the revisions walker
            constructor
        kwargs (dict): keyword arguments to pass to the revisions walker
            constructor

    """
    # first check if the provided revision is valid
    lookup_revision(rev_start)
    return _RevisionsWalkerProxy(rev_walker_type, rev_start, *args, **kwargs)


def lookup_object(object_type: str, object_id: str) -> Dict[str, Any]:
    """
    Utility function for looking up an object in the archive by its type
    and id.

    Args:
        object_type (str): the type of object to lookup, either *content*,
            *directory*, *release*, *revision* or *snapshot*
        object_id (str): the *sha1_git* checksum identifier in hexadecimal
            form of the object to lookup

    Returns:
        Dict[str, Any]: A dictionary describing the object or a list of
        dictionary for the directory object type.

    Raises:
        swh.web.common.exc.NotFoundExc: if the object could not be found in
            the archive
        BadInputExc: if the object identifier is invalid
    """
    if object_type == CONTENT:
        return lookup_content(f"sha1_git:{object_id}")
    elif object_type == DIRECTORY:
        return {"id": object_id, "content": list(lookup_directory(object_id))}
    elif object_type == RELEASE:
        return lookup_release(object_id)
    elif object_type == REVISION:
        return lookup_revision(object_id)
    elif object_type == SNAPSHOT:
        return lookup_snapshot(object_id)

    raise BadInputExc(
        (
            "Invalid swh object type! Valid types are "
            f"{CONTENT}, {DIRECTORY}, {RELEASE} "
            f"{REVISION} or {SNAPSHOT}."
        )
    )


def lookup_missing_hashes(grouped_pids: Dict[str, List[bytes]]) -> Set[str]:
    """Lookup missing Software Heritage persistent identifier hash, using
    batch processing.

    Args:
        A dictionary with:
        keys: persistent identifier type
        values: list(bytes) persistent identifier hash
    Returns:
        A set(hexadecimal) of the hashes not found in the storage
    """
    missing_hashes = []

    for obj_type, obj_ids in grouped_pids.items():
        if obj_type == CONTENT:
            missing_hashes.append(storage.content_missing_per_sha1_git(obj_ids))
        elif obj_type == DIRECTORY:
            missing_hashes.append(storage.directory_missing(obj_ids))
        elif obj_type == REVISION:
            missing_hashes.append(storage.revision_missing(obj_ids))
        elif obj_type == RELEASE:
            missing_hashes.append(storage.directory_missing(obj_ids))
        elif obj_type == SNAPSHOT:
            missing_hashes.append(storage.directory_missing(obj_ids))

    missing = set(
        map(lambda x: hashutil.hash_to_hex(x), itertools.chain(*missing_hashes))
    )

    return missing



# Copyright (C) 2018-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

# Utility module for browsing the archive in a snapshot context.

from collections import defaultdict
from copy import copy
from typing import Any, Dict, List, Optional, Union, Tuple


from django.core.cache import cache
from django.shortcuts import render
from django.template.defaultfilters import filesizeformat
from django.utils.html import escape
import sentry_sdk

from swh.model.identifiers import (
    persistent_identifier,
    snapshot_identifier,
    CONTENT,
    DIRECTORY,
    REVISION,
    RELEASE,
    SNAPSHOT,
)

from swh.web.browse.utils import (
    get_directory_entries,
    gen_directory_link,
    gen_revision_link,
    request_content,
    gen_content_link,
    prepare_content_for_display,
    content_display_max_size,
    format_log_entries,
    gen_revision_log_link,
    gen_release_link,
    get_readme_to_display,
    gen_snapshot_link,
)

from swh.web.common import service, highlightjs
from swh.web.common.exc import handle_view_exception, NotFoundExc, BadInputExc
from swh.web.common.identifiers import get_swhids_info
from swh.web.common.origin_visits import get_origin_visit
from swh.web.common.typing import (
    OriginInfo,
    SnapshotBranchInfo,
    SnapshotReleaseInfo,
    SnapshotContext,
    ContentMetadata,
    DirectoryMetadata,
    SWHObjectInfo,
)
from swh.web.common.utils import (
    reverse,
    gen_path_info,
    format_utc_iso_date,
    swh_object_icons,
)
from swh.web.config import get_config


_empty_snapshot_id = snapshot_identifier({"branches": {}})


def _get_branch(branches, branch_name, snapshot_id):
    """
    Utility function to get a specific branch from a branches list.
    Its purpose is to get the default HEAD branch as some software origin
    (e.g those with svn type) does not have it. In that latter case, check
    if there is a master branch instead and returns it.
    """
    filtered_branches = [b for b in branches if b["name"] == branch_name]
    if filtered_branches:
        return filtered_branches[0]
    elif branch_name == "HEAD":
        filtered_branches = [b for b in branches if b["name"].endswith("master")]
        if filtered_branches:
            return filtered_branches[0]
        elif branches:
            return branches[0]
    else:
        # case where a large branches list has been truncated
        snp = service.lookup_snapshot(
            snapshot_id,
            branches_from=branch_name,
            branches_count=1,
            target_types=["revision", "alias"],
        )
        snp_branch, _ = process_snapshot_branches(snp)
        if snp_branch and snp_branch[0]["name"] == branch_name:
            branches.append(snp_branch[0])
            return snp_branch[0]


def _get_release(releases, release_name, snapshot_id):
    """
    Utility function to get a specific release from a releases list.
    Returns None if the release can not be found in the list.
    """
    filtered_releases = [r for r in releases if r["name"] == release_name]
    if filtered_releases:
        return filtered_releases[0]
    else:
        # case where a large branches list has been truncated
        try:
            # git origins have specific branches for releases
            snp = service.lookup_snapshot(
                snapshot_id,
                branches_from=f"refs/tags/{release_name}",
                branches_count=1,
                target_types=["release"],
            )
        except NotFoundExc:
            snp = service.lookup_snapshot(
                snapshot_id,
                branches_from=release_name,
                branches_count=1,
                target_types=["release"],
            )
        _, snp_release = process_snapshot_branches(snp)
        if snp_release and snp_release[0]["name"] == release_name:
            releases.append(snp_release[0])
            return snp_release[0]


def _branch_not_found(
    branch_type, branch, snapshot_id, snapshot_sizes, origin_info, timestamp, visit_id
):
    """
    Utility function to raise an exception when a specified branch/release
    can not be found.
    """
    if branch_type == "branch":
        branch_type = "Branch"
        branch_type_plural = "branches"
        target_type = "revision"
    else:
        branch_type = "Release"
        branch_type_plural = "releases"
        target_type = "release"

    if snapshot_id and snapshot_sizes[target_type] == 0:
        msg = "Snapshot with id %s has an empty list" " of %s!" % (
            snapshot_id,
            branch_type_plural,
        )
    elif snapshot_id:
        msg = "%s %s for snapshot with id %s" " not found!" % (
            branch_type,
            branch,
            snapshot_id,
        )
    elif visit_id and snapshot_sizes[target_type] == 0:
        msg = (
            "Origin with url %s"
            " for visit with id %s has an empty list"
            " of %s!" % (origin_info["url"], visit_id, branch_type_plural)
        )
    elif visit_id:
        msg = (
            "%s %s associated to visit with"
            " id %s for origin with url %s"
            " not found!" % (branch_type, branch, visit_id, origin_info["url"])
        )
    elif snapshot_sizes[target_type] == 0:
        msg = (
            "Origin with url %s"
            " for visit with timestamp %s has an empty list"
            " of %s!" % (origin_info["url"], timestamp, branch_type_plural)
        )
    else:
        msg = (
            "%s %s associated to visit with"
            " timestamp %s for origin with "
            "url %s not found!" % (branch_type, branch, timestamp, origin_info["url"])
        )
    raise NotFoundExc(escape(msg))


def process_snapshot_branches(
    snapshot: Dict[str, Any]
) -> Tuple[List[SnapshotBranchInfo], List[SnapshotReleaseInfo]]:
    """
    Process a dictionary describing snapshot branches: extract those
    targeting revisions and releases, put them in two different lists,
    then sort those lists in lexicographical order of the branches' names.

    Args:
        snapshot: A dict describing a snapshot as returned for instance by
            :func:`swh.web.common.service.lookup_snapshot`

    Returns:
        A tuple whose first member is the sorted list of branches
        targeting revisions and second member the sorted list of branches
        targeting releases
    """
    snapshot_branches = snapshot["branches"]
    branches: Dict[str, SnapshotBranchInfo] = {}
    branch_aliases: Dict[str, str] = {}
    releases: Dict[str, SnapshotReleaseInfo] = {}
    revision_to_branch = defaultdict(set)
    revision_to_release = defaultdict(set)
    release_to_branch = defaultdict(set)
    for branch_name, target in snapshot_branches.items():
        if not target:
            # FIXME: display branches with an unknown target anyway
            continue
        target_id = target["target"]
        target_type = target["target_type"]
        if target_type == "revision":
            branches[branch_name] = SnapshotBranchInfo(
                name=branch_name,
                revision=target_id,
                date=None,
                directory=None,
                message=None,
                url=None,
            )
            revision_to_branch[target_id].add(branch_name)
        elif target_type == "release":
            release_to_branch[target_id].add(branch_name)
        elif target_type == "alias":
            branch_aliases[branch_name] = target_id
        # FIXME: handle pointers to other object types

    def _add_release_info(branch, release):
        releases[branch] = SnapshotReleaseInfo(
            name=release["name"],
            branch_name=branch,
            date=format_utc_iso_date(release["date"]),
            directory=None,
            id=release["id"],
            message=release["message"],
            target_type=release["target_type"],
            target=release["target"],
            url=None,
        )

    def _add_branch_info(branch, revision):
        branches[branch] = SnapshotBranchInfo(
            name=branch,
            revision=revision["id"],
            directory=revision["directory"],
            date=format_utc_iso_date(revision["date"]),
            message=revision["message"],
            url=None,
        )

    releases_info = service.lookup_release_multiple(release_to_branch.keys())
    for release in releases_info:
        branches_to_update = release_to_branch[release["id"]]
        for branch in branches_to_update:
            _add_release_info(branch, release)
        if release["target_type"] == "revision":
            revision_to_release[release["target"]].update(branches_to_update)

    revisions = service.lookup_revision_multiple(
        set(revision_to_branch.keys()) | set(revision_to_release.keys())
    )

    for revision in revisions:
        if not revision:
            continue
        for branch in revision_to_branch[revision["id"]]:
            _add_branch_info(branch, revision)
        for release in revision_to_release[revision["id"]]:
            releases[release]["directory"] = revision["directory"]

    for branch_alias, branch_target in branch_aliases.items():
        if branch_target in branches:
            branches[branch_alias] = copy(branches[branch_target])
        else:
            snp = service.lookup_snapshot(
                snapshot["id"], branches_from=branch_target, branches_count=1
            )
            if snp and branch_target in snp["branches"]:

                if snp["branches"][branch_target] is None:
                    continue

                target_type = snp["branches"][branch_target]["target_type"]
                target = snp["branches"][branch_target]["target"]
                if target_type == "revision":
                    branches[branch_alias] = snp["branches"][branch_target]
                    revision = service.lookup_revision(target)
                    _add_branch_info(branch_alias, revision)
                elif target_type == "release":
                    release = service.lookup_release(target)
                    _add_release_info(branch_alias, release)

        if branch_alias in branches:
            branches[branch_alias]["name"] = branch_alias

    ret_branches = list(sorted(branches.values(), key=lambda b: b["name"]))
    ret_releases = list(sorted(releases.values(), key=lambda b: b["name"]))

    return ret_branches, ret_releases


def get_snapshot_content(
    snapshot_id: str,
) -> Tuple[List[SnapshotBranchInfo], List[SnapshotReleaseInfo]]:
    """Returns the lists of branches and releases
    associated to a swh snapshot.
    That list is put in  cache in order to speedup the navigation
    in the swh-web/browse ui.

    .. warning:: At most 1000 branches contained in the snapshot
        will be returned for performance reasons.

    Args:
        snapshot_id: hexadecimal representation of the snapshot identifier

    Returns:
        A tuple with two members. The first one is a list of dict describing
        the snapshot branches. The second one is a list of dict describing the
        snapshot releases.

    Raises:
        NotFoundExc if the snapshot does not exist
    """
    cache_entry_id = "swh_snapshot_%s" % snapshot_id
    cache_entry = cache.get(cache_entry_id)

    if cache_entry:
        return cache_entry["branches"], cache_entry["releases"]

    branches: List[SnapshotBranchInfo] = []
    releases: List[SnapshotReleaseInfo] = []

    snapshot_content_max_size = get_config()["snapshot_content_max_size"]

    if snapshot_id:
        snapshot = service.lookup_snapshot(
            snapshot_id, branches_count=snapshot_content_max_size
        )
        branches, releases = process_snapshot_branches(snapshot)

    cache.set(cache_entry_id, {"branches": branches, "releases": releases,})

    return branches, releases


def get_origin_visit_snapshot(
    origin_info: OriginInfo,
    visit_ts: Optional[Union[int, str]] = None,
    visit_id: Optional[int] = None,
    snapshot_id: Optional[str] = None,
) -> Tuple[List[SnapshotBranchInfo], List[SnapshotReleaseInfo]]:
    """Returns the lists of branches and releases associated to an origin for
    a given visit.

    The visit is expressed by either:

        * a snapshot identifier
        * a timestamp, if no visit with that exact timestamp is found,
          the closest one from the provided timestamp will be used.

    If no visit parameter is provided, it returns the list of branches
    found for the latest visit.

    That list is put in  cache in order to speedup the navigation
    in the swh-web/browse ui.

    .. warning:: At most 1000 branches contained in the snapshot
        will be returned for performance reasons.

    Args:
        origin_info: a dict filled with origin information
        visit_ts: an ISO date string or Unix timestamp to parse
        visit_id: visit id for disambiguation in case several visits have
            the same timestamp
        snapshot_id: if provided, visit associated to the snapshot will be processed

    Returns:
        A tuple with two members. The first one is a list of dict describing
        the origin branches for the given visit.
        The second one is a list of dict describing the origin releases
        for the given visit.

    Raises:
        NotFoundExc if the origin or its visit are not found
    """

    visit_info = get_origin_visit(origin_info, visit_ts, visit_id, snapshot_id)

    return get_snapshot_content(visit_info["snapshot"])


def get_snapshot_context(
    snapshot_id: Optional[str] = None,
    origin_url: Optional[str] = None,
    timestamp: Optional[str] = None,
    visit_id: Optional[int] = None,
    branch_name: Optional[str] = None,
    release_name: Optional[str] = None,
    revision_id: Optional[str] = None,
    path: Optional[str] = None,
    browse_context: str = "directory",
) -> SnapshotContext:
    """
    Utility function to compute relevant information when navigating
    the archive in a snapshot context. The snapshot is either
    referenced by its id or it will be retrieved from an origin visit.

    Args:
        snapshot_id: hexadecimal representation of a snapshot identifier
        origin_url: an origin_url
        timestamp: a datetime string for retrieving the closest
            visit of the origin
        visit_id: optional visit id for disambiguation in case
            of several visits with the same timestamp
        branch_name: optional branch name set when browsing the snapshot in
            that scope (will default to "HEAD" if not provided)
        release_name: optional release name set when browsing the snapshot in
            that scope
        revision_id: optional revision identifier set when browsing the snapshot in
            that scope
        path: optional path of the object currently browsed in the snapshot
        browse_context: indicates which type of object is currently browsed

    Returns:
        A dict filled with snapshot context information.

    Raises:
        swh.web.common.exc.NotFoundExc: if no snapshot is found for the visit
            of an origin.
    """
    assert origin_url is not None or snapshot_id is not None
    origin_info = None
    visit_info = None
    url_args = {}
    query_params: Dict[str, Any] = {}
    origin_visits_url = None

    if origin_url:

        if visit_id is not None:
            query_params["visit_id"] = visit_id
        elif snapshot_id is not None:
            query_params["snapshot"] = snapshot_id

        origin_info = service.lookup_origin({"url": origin_url})

        visit_info = get_origin_visit(origin_info, timestamp, visit_id, snapshot_id)
        formatted_date = format_utc_iso_date(visit_info["date"])
        visit_info["formatted_date"] = formatted_date
        snapshot_id = visit_info["snapshot"]

        if not snapshot_id:
            raise NotFoundExc(
                "No snapshot associated to the visit of origin "
                "%s on %s" % (escape(origin_url), formatted_date)
            )

        # provided timestamp is not necessarily equals to the one
        # of the retrieved visit, so get the exact one in order
        # to use it in the urls generated below
        if timestamp:
            timestamp = visit_info["date"]

        branches, releases = get_origin_visit_snapshot(
            origin_info, timestamp, visit_id, snapshot_id
        )

        query_params["origin_url"] = origin_info["url"]

        origin_visits_url = reverse(
            "browse-origin-visits", query_params={"origin_url": origin_info["url"]}
        )

        if timestamp is not None:
            query_params["timestamp"] = format_utc_iso_date(
                timestamp, "%Y-%m-%dT%H:%M:%SZ"
            )

        visit_url = reverse("browse-origin-directory", query_params=query_params)
        visit_info["url"] = visit_url

        branches_url = reverse("browse-origin-branches", query_params=query_params)

        releases_url = reverse("browse-origin-releases", query_params=query_params)
    else:
        assert snapshot_id is not None
        branches, releases = get_snapshot_content(snapshot_id)
        url_args = {"snapshot_id": snapshot_id}
        branches_url = reverse("browse-snapshot-branches", url_args=url_args)

        releases_url = reverse("browse-snapshot-releases", url_args=url_args)

    releases = list(reversed(releases))

    snapshot_sizes = service.lookup_snapshot_sizes(snapshot_id)

    is_empty = sum(snapshot_sizes.values()) == 0

    swh_snp_id = persistent_identifier("snapshot", snapshot_id)

    if visit_info:
        timestamp = format_utc_iso_date(visit_info["date"])

    if origin_info:
        browse_view_name = f"browse-origin-{browse_context}"
    else:
        browse_view_name = f"browse-snapshot-{browse_context}"

    release_id = None
    root_directory = None

    snapshot_total_size = sum(snapshot_sizes.values())

    if path is not None:
        query_params["path"] = path

    if snapshot_total_size and revision_id is not None:
        revision = service.lookup_revision(revision_id)
        root_directory = revision["directory"]
        branches.append(
            SnapshotBranchInfo(
                name=revision_id,
                revision=revision_id,
                directory=root_directory,
                date=revision["date"],
                message=revision["message"],
                url=None,
            )
        )
        branch_name = revision_id
        query_params["revision"] = revision_id
    elif snapshot_total_size and release_name:
        release = _get_release(releases, release_name, snapshot_id)
        try:
            root_directory = release["directory"]
            revision_id = release["target"]
            release_id = release["id"]
            query_params["release"] = release_name
        except Exception as exc:
            sentry_sdk.capture_exception(exc)
            _branch_not_found(
                "release",
                release_name,
                snapshot_id,
                snapshot_sizes,
                origin_info,
                timestamp,
                visit_id,
            )
    elif snapshot_total_size:
        if branch_name:
            query_params["branch"] = branch_name
        branch = _get_branch(branches, branch_name or "HEAD", snapshot_id)
        try:
            branch_name = branch["name"]
            revision_id = branch["revision"]
            root_directory = branch["directory"]
        except Exception as exc:
            sentry_sdk.capture_exception(exc)
            _branch_not_found(
                "branch",
                branch_name,
                snapshot_id,
                snapshot_sizes,
                origin_info,
                timestamp,
                visit_id,
            )

    for b in branches:
        branch_query_params = dict(query_params)
        branch_query_params.pop("release", None)
        if b["name"] != b["revision"]:
            branch_query_params.pop("revision", None)
            branch_query_params["branch"] = b["name"]
        b["url"] = reverse(
            browse_view_name, url_args=url_args, query_params=branch_query_params
        )

    for r in releases:
        release_query_params = dict(query_params)
        release_query_params.pop("branch", None)
        release_query_params.pop("revision", None)
        release_query_params["release"] = r["name"]
        r["url"] = reverse(
            browse_view_name, url_args=url_args, query_params=release_query_params,
        )

    return SnapshotContext(
        branch=branch_name,
        branches=branches,
        branches_url=branches_url,
        is_empty=is_empty,
        origin_info=origin_info,
        origin_visits_url=origin_visits_url,
        release=release_name,
        release_id=release_id,
        query_params=query_params,
        releases=releases,
        releases_url=releases_url,
        revision_id=revision_id,
        root_directory=root_directory,
        snapshot_id=snapshot_id,
        snapshot_sizes=snapshot_sizes,
        snapshot_swhid=swh_snp_id,
        url_args=url_args,
        visit_info=visit_info,
    )


def _build_breadcrumbs(snapshot_context: SnapshotContext, path: str):
    origin_info = snapshot_context["origin_info"]
    url_args = snapshot_context["url_args"]
    query_params = dict(snapshot_context["query_params"])
    root_directory = snapshot_context["root_directory"]

    path_info = gen_path_info(path)

    if origin_info:
        browse_view_name = "browse-origin-directory"
    else:
        browse_view_name = "browse-snapshot-directory"

    breadcrumbs = []
    if root_directory:
        query_params.pop("path", None)
        breadcrumbs.append(
            {
                "name": root_directory[:7],
                "url": reverse(
                    browse_view_name, url_args=url_args, query_params=query_params
                ),
            }
        )
    for pi in path_info:
        query_params["path"] = pi["path"]
        breadcrumbs.append(
            {
                "name": pi["name"],
                "url": reverse(
                    browse_view_name, url_args=url_args, query_params=query_params
                ),
            }
        )
    return breadcrumbs


def _check_origin_url(snapshot_id, origin_url):
    if snapshot_id is None and origin_url is None:
        raise BadInputExc("An origin URL must be provided as query parameter.")


def browse_snapshot_directory(
    request, snapshot_id=None, origin_url=None, timestamp=None, path=None
):
    """
    Django view implementation for browsing a directory in a snapshot context.
    """
    try:

        _check_origin_url(snapshot_id, origin_url)

        snapshot_context = get_snapshot_context(
            snapshot_id=snapshot_id,
            origin_url=origin_url,
            timestamp=timestamp,
            visit_id=request.GET.get("visit_id"),
            path=path,
            browse_context="directory",
            branch_name=request.GET.get("branch"),
            release_name=request.GET.get("release"),
            revision_id=request.GET.get("revision"),
        )

        root_directory = snapshot_context["root_directory"]
        sha1_git = root_directory
        if root_directory and path:
            dir_info = service.lookup_directory_with_path(root_directory, path)
            sha1_git = dir_info["target"]

        dirs = []
        files = []
        if sha1_git:
            dirs, files = get_directory_entries(sha1_git)

    except Exception as exc:
        return handle_view_exception(request, exc)

    origin_info = snapshot_context["origin_info"]
    visit_info = snapshot_context["visit_info"]
    url_args = snapshot_context["url_args"]
    query_params = dict(snapshot_context["query_params"])
    revision_id = snapshot_context["revision_id"]
    snapshot_id = snapshot_context["snapshot_id"]

    if origin_info:
        browse_view_name = "browse-origin-directory"
    else:
        browse_view_name = "browse-snapshot-directory"

    breadcrumbs = _build_breadcrumbs(snapshot_context, path)

    path = "" if path is None else (path + "/")

    for d in dirs:
        if d["type"] == "rev":
            d["url"] = reverse("browse-revision", url_args={"sha1_git": d["target"]})
        else:
            query_params["path"] = path + d["name"]
            d["url"] = reverse(
                browse_view_name, url_args=url_args, query_params=query_params
            )

    sum_file_sizes = 0

    readmes = {}

    if origin_info:
        browse_view_name = "browse-origin-content"
    else:
        browse_view_name = "browse-snapshot-content"

    for f in files:
        query_params["path"] = path + f["name"]
        f["url"] = reverse(
            browse_view_name, url_args=url_args, query_params=query_params
        )
        if f["length"] is not None:
            sum_file_sizes += f["length"]
            f["length"] = filesizeformat(f["length"])
        if f["name"].lower().startswith("readme"):
            readmes[f["name"]] = f["checksums"]["sha1"]

    readme_name, readme_url, readme_html = get_readme_to_display(readmes)

    if origin_info:
        browse_view_name = "browse-origin-log"
    else:
        browse_view_name = "browse-snapshot-log"

    history_url = None
    if snapshot_id != _empty_snapshot_id:
        query_params.pop("path", None)
        history_url = reverse(
            browse_view_name, url_args=url_args, query_params=query_params
        )

    nb_files = None
    nb_dirs = None
    dir_path = None
    if root_directory:
        nb_files = len(files)
        nb_dirs = len(dirs)
        sum_file_sizes = filesizeformat(sum_file_sizes)
        dir_path = "/" + path

    browse_dir_link = gen_directory_link(sha1_git)

    browse_rev_link = gen_revision_link(revision_id)
    browse_snp_link = gen_snapshot_link(snapshot_id)

    revision_found = True
    if sha1_git is None and revision_id is not None:
        try:
            service.lookup_revision(revision_id)
        except NotFoundExc:
            revision_found = False

    swh_objects = [
        SWHObjectInfo(object_type=DIRECTORY, object_id=sha1_git),
        SWHObjectInfo(object_type=REVISION, object_id=revision_id),
        SWHObjectInfo(object_type=SNAPSHOT, object_id=snapshot_id),
    ]

    visit_date = None
    visit_type = None
    if visit_info:
        visit_date = format_utc_iso_date(visit_info["date"])
        visit_type = visit_info["type"]

    release_id = snapshot_context["release_id"]
    browse_rel_link = None
    if release_id:
        swh_objects.append(SWHObjectInfo(object_type=RELEASE, object_id=release_id))
        browse_rel_link = gen_release_link(release_id)

    dir_metadata = DirectoryMetadata(
        object_type=DIRECTORY,
        object_id=sha1_git,
        directory=sha1_git,
        directory_url=browse_dir_link,
        nb_files=nb_files,
        nb_dirs=nb_dirs,
        sum_file_sizes=sum_file_sizes,
        root_directory=root_directory,
        path=dir_path,
        revision=revision_id,
        revision_found=revision_found,
        revision_url=browse_rev_link,
        release=release_id,
        release_url=browse_rel_link,
        snapshot=snapshot_id,
        snapshot_url=browse_snp_link,
        origin_url=origin_url,
        visit_date=visit_date,
        visit_type=visit_type,
    )

    vault_cooking = {
        "directory_context": True,
        "directory_id": sha1_git,
        "revision_context": True,
        "revision_id": revision_id,
    }

    swhids_info = get_swhids_info(swh_objects, snapshot_context, dir_metadata)

    dir_path = "/".join([bc["name"] for bc in breadcrumbs]) + "/"
    context_found = "snapshot: %s" % snapshot_context["snapshot_id"]
    if origin_info:
        context_found = "origin: %s" % origin_info["url"]
    heading = "Directory - %s - %s - %s" % (
        dir_path,
        snapshot_context["branch"],
        context_found,
    )

    top_right_link = None
    if not snapshot_context["is_empty"]:
        top_right_link = {
            "url": history_url,
            "icon": swh_object_icons["revisions history"],
            "text": "History",
        }

    return render(
        request,
        "browse/directory.html",
        {
            "heading": heading,
            "swh_object_name": "Directory",
            "swh_object_metadata": dir_metadata,
            "dirs": dirs,
            "files": files,
            "breadcrumbs": breadcrumbs if root_directory else [],
            "top_right_link": top_right_link,
            "readme_name": readme_name,
            "readme_url": readme_url,
            "readme_html": readme_html,
            "snapshot_context": snapshot_context,
            "vault_cooking": vault_cooking,
            "show_actions_menu": True,
            "swhids_info": swhids_info,
        },
    )


def browse_snapshot_content(
    request,
    snapshot_id=None,
    origin_url=None,
    timestamp=None,
    path=None,
    selected_language=None,
):
    """
    Django view implementation for browsing a content in a snapshot context.
    """
    try:

        _check_origin_url(snapshot_id, origin_url)

        if path is None:
            raise BadInputExc("The path of a content must be given as query parameter.")

        snapshot_context = get_snapshot_context(
            snapshot_id=snapshot_id,
            origin_url=origin_url,
            timestamp=timestamp,
            visit_id=request.GET.get("visit_id"),
            path=path,
            browse_context="content",
            branch_name=request.GET.get("branch"),
            release_name=request.GET.get("release"),
            revision_id=request.GET.get("revision"),
        )

        root_directory = snapshot_context["root_directory"]
        sha1_git = None
        query_string = None
        content_data = {}
        directory_id = None
        split_path = path.split("/")
        filename = split_path[-1]
        filepath = path[: -len(filename)]
        if root_directory:
            content_info = service.lookup_directory_with_path(root_directory, path)
            sha1_git = content_info["target"]
            query_string = "sha1_git:" + sha1_git
            content_data = request_content(query_string, raise_if_unavailable=False)

            if filepath:
                dir_info = service.lookup_directory_with_path(root_directory, filepath)
                directory_id = dir_info["target"]
            else:
                directory_id = root_directory

    except Exception as exc:
        return handle_view_exception(request, exc)

    revision_id = snapshot_context["revision_id"]
    origin_info = snapshot_context["origin_info"]
    visit_info = snapshot_context["visit_info"]
    snapshot_id = snapshot_context["snapshot_id"]

    if content_data.get("raw_data") is not None:
        content_display_data = prepare_content_for_display(
            content_data["raw_data"], content_data["mimetype"], path
        )
        content_data.update(content_display_data)

    # Override language with user-selected language
    if selected_language is not None:
        content_data["language"] = selected_language

    available_languages = None

    if content_data.get("mimetype") is not None and "text/" in content_data["mimetype"]:
        available_languages = highlightjs.get_supported_languages()

    breadcrumbs = _build_breadcrumbs(snapshot_context, filepath)

    breadcrumbs.append({"name": filename, "url": None})

    browse_content_link = gen_content_link(sha1_git)

    content_raw_url = None
    if query_string:
        content_raw_url = reverse(
            "browse-content-raw",
            url_args={"query_string": query_string},
            query_params={"filename": filename},
        )

    browse_rev_link = gen_revision_link(revision_id)

    browse_dir_link = gen_directory_link(directory_id)

    content_checksums = content_data.get("checksums", {})

    swh_objects = [
        SWHObjectInfo(object_type=CONTENT, object_id=content_checksums.get("sha1_git")),
        SWHObjectInfo(object_type=DIRECTORY, object_id=directory_id),
        SWHObjectInfo(object_type=REVISION, object_id=revision_id),
        SWHObjectInfo(object_type=SNAPSHOT, object_id=snapshot_id),
    ]

    visit_date = None
    visit_type = None
    if visit_info:
        visit_date = format_utc_iso_date(visit_info["date"])
        visit_type = visit_info["type"]

    release_id = snapshot_context["release_id"]
    browse_rel_link = None
    if release_id:
        swh_objects.append(SWHObjectInfo(object_type=RELEASE, object_id=release_id))
        browse_rel_link = gen_release_link(release_id)

    content_metadata = ContentMetadata(
        object_type=CONTENT,
        object_id=content_checksums.get("sha1_git"),
        sha1=content_checksums.get("sha1"),
        sha1_git=content_checksums.get("sha1_git"),
        sha256=content_checksums.get("sha256"),
        blake2s256=content_checksums.get("blake2s256"),
        content_url=browse_content_link,
        mimetype=content_data.get("mimetype"),
        encoding=content_data.get("encoding"),
        size=filesizeformat(content_data.get("length", 0)),
        language=content_data.get("language"),
        licenses=content_data.get("licenses"),
        root_directory=root_directory,
        path=f"/{filepath}",
        filename=filename,
        directory=directory_id,
        directory_url=browse_dir_link,
        revision=revision_id,
        revision_url=browse_rev_link,
        release=release_id,
        release_url=browse_rel_link,
        snapshot=snapshot_id,
        snapshot_url=gen_snapshot_link(snapshot_id),
        origin_url=origin_url,
        visit_date=visit_date,
        visit_type=visit_type,
    )

    swhids_info = get_swhids_info(swh_objects, snapshot_context, content_metadata)

    content_path = "/".join([bc["name"] for bc in breadcrumbs])
    context_found = "snapshot: %s" % snapshot_context["snapshot_id"]
    if origin_info:
        context_found = "origin: %s" % origin_info["url"]
    heading = "Content - %s - %s - %s" % (
        content_path,
        snapshot_context["branch"],
        context_found,
    )

    top_right_link = None
    if not snapshot_context["is_empty"]:
        top_right_link = {
            "url": content_raw_url,
            "icon": swh_object_icons["content"],
            "text": "Raw File",
        }

    return render(
        request,
        "browse/content.html",
        {
            "heading": heading,
            "swh_object_name": "Content",
            "swh_object_metadata": content_metadata,
            "content": content_data.get("content_data"),
            "content_size": content_data.get("length"),
            "max_content_size": content_display_max_size,
            "filename": filename,
            "encoding": content_data.get("encoding"),
            "mimetype": content_data.get("mimetype"),
            "language": content_data.get("language"),
            "available_languages": available_languages,
            "breadcrumbs": breadcrumbs if root_directory else [],
            "top_right_link": top_right_link,
            "snapshot_context": snapshot_context,
            "vault_cooking": None,
            "show_actions_menu": True,
            "swhids_info": swhids_info,
            "error_code": content_data.get("error_code"),
            "error_message": content_data.get("error_message"),
            "error_description": content_data.get("error_description"),
        },
        status=content_data.get("error_code", 200),
    )


PER_PAGE = 100


def browse_snapshot_log(request, snapshot_id=None, origin_url=None, timestamp=None):
    """
    Django view implementation for browsing a revision history in a
    snapshot context.
    """
    try:

        _check_origin_url(snapshot_id, origin_url)

        snapshot_context = get_snapshot_context(
            snapshot_id=snapshot_id,
            origin_url=origin_url,
            timestamp=timestamp,
            visit_id=request.GET.get("visit_id"),
            browse_context="log",
            branch_name=request.GET.get("branch"),
            release_name=request.GET.get("release"),
            revision_id=request.GET.get("revision"),
        )

        revision_id = snapshot_context["revision_id"]

        per_page = int(request.GET.get("per_page", PER_PAGE))
        offset = int(request.GET.get("offset", 0))
        revs_ordering = request.GET.get("revs_ordering", "committer_date")
        session_key = "rev_%s_log_ordering_%s" % (revision_id, revs_ordering)
        rev_log_session = request.session.get(session_key, None)
        rev_log = []
        revs_walker_state = None
        if rev_log_session:
            rev_log = rev_log_session["rev_log"]
            revs_walker_state = rev_log_session["revs_walker_state"]

        if len(rev_log) < offset + per_page:
            revs_walker = service.get_revisions_walker(
                revs_ordering,
                revision_id,
                max_revs=offset + per_page + 1,
                state=revs_walker_state,
            )
            rev_log += [rev["id"] for rev in revs_walker]
            revs_walker_state = revs_walker.export_state()

        revs = rev_log[offset : offset + per_page]
        revision_log = service.lookup_revision_multiple(revs)

        request.session[session_key] = {
            "rev_log": rev_log,
            "revs_walker_state": revs_walker_state,
        }

    except Exception as exc:
        return handle_view_exception(request, exc)

    origin_info = snapshot_context["origin_info"]
    visit_info = snapshot_context["visit_info"]
    url_args = snapshot_context["url_args"]
    query_params = snapshot_context["query_params"]
    snapshot_id = snapshot_context["snapshot_id"]

    query_params["per_page"] = per_page
    revs_ordering = request.GET.get("revs_ordering", "")
    query_params["revs_ordering"] = revs_ordering

    if origin_info:
        browse_view_name = "browse-origin-log"
    else:
        browse_view_name = "browse-snapshot-log"

    prev_log_url = None
    if len(rev_log) > offset + per_page:
        query_params["offset"] = offset + per_page
        prev_log_url = reverse(
            browse_view_name, url_args=url_args, query_params=query_params
        )

    next_log_url = None
    if offset != 0:
        query_params["offset"] = offset - per_page
        next_log_url = reverse(
            browse_view_name, url_args=url_args, query_params=query_params
        )

    revision_log_data = format_log_entries(revision_log, per_page, snapshot_context)

    browse_rev_link = gen_revision_link(revision_id)

    browse_log_link = gen_revision_log_link(revision_id)

    browse_snp_link = gen_snapshot_link(snapshot_id)

    revision_metadata = {
        "context-independent revision": browse_rev_link,
        "context-independent revision history": browse_log_link,
        "context-independent snapshot": browse_snp_link,
        "snapshot": snapshot_id,
    }

    if origin_info:
        revision_metadata["origin url"] = origin_info["url"]
        revision_metadata["origin visit date"] = format_utc_iso_date(visit_info["date"])
        revision_metadata["origin visit type"] = visit_info["type"]

    swh_objects = [
        SWHObjectInfo(object_type=REVISION, object_id=revision_id),
        SWHObjectInfo(object_type=SNAPSHOT, object_id=snapshot_id),
    ]

    release_id = snapshot_context["release_id"]
    if release_id:
        swh_objects.append(SWHObjectInfo(object_type=RELEASE, object_id=release_id))
        browse_rel_link = gen_release_link(release_id)
        revision_metadata["release"] = release_id
        revision_metadata["context-independent release"] = browse_rel_link

    swhids_info = get_swhids_info(swh_objects, snapshot_context)

    context_found = "snapshot: %s" % snapshot_context["snapshot_id"]
    if origin_info:
        context_found = "origin: %s" % origin_info["url"]
    heading = "Revision history - %s - %s" % (snapshot_context["branch"], context_found)

    return render(
        request,
        "browse/revision-log.html",
        {
            "heading": heading,
            "swh_object_name": "Revisions history",
            "swh_object_metadata": revision_metadata,
            "revision_log": revision_log_data,
            "revs_ordering": revs_ordering,
            "next_log_url": next_log_url,
            "prev_log_url": prev_log_url,
            "breadcrumbs": None,
            "top_right_link": None,
            "snapshot_context": snapshot_context,
            "vault_cooking": None,
            "show_actions_menu": True,
            "swhids_info": swhids_info,
        },
    )


def browse_snapshot_branches(
    request, snapshot_id=None, origin_url=None, timestamp=None
):
    """
    Django view implementation for browsing a list of branches in a snapshot
    context.
    """
    try:

        _check_origin_url(snapshot_id, origin_url)

        snapshot_context = get_snapshot_context(
            snapshot_id=snapshot_id,
            origin_url=origin_url,
            timestamp=timestamp,
            visit_id=request.GET.get("visit_id"),
        )

        branches_bc = request.GET.get("branches_breadcrumbs", "")
        branches_bc = branches_bc.split(",") if branches_bc else []
        branches_from = branches_bc[-1] if branches_bc else ""

        origin_info = snapshot_context["origin_info"]
        url_args = snapshot_context["url_args"]
        query_params = snapshot_context["query_params"]

        if origin_info:
            browse_view_name = "browse-origin-directory"
        else:
            browse_view_name = "browse-snapshot-directory"

        snapshot = service.lookup_snapshot(
            snapshot_context["snapshot_id"],
            branches_from,
            PER_PAGE + 1,
            target_types=["revision", "alias"],
        )

        displayed_branches, _ = process_snapshot_branches(snapshot)

    except Exception as exc:
        return handle_view_exception(request, exc)

    for branch in displayed_branches:
        rev_query_params = {}
        if origin_info:
            rev_query_params["origin_url"] = origin_info["url"]

        revision_url = reverse(
            "browse-revision",
            url_args={"sha1_git": branch["revision"]},
            query_params=query_params,
        )

        query_params["branch"] = branch["name"]
        directory_url = reverse(
            browse_view_name, url_args=url_args, query_params=query_params
        )
        del query_params["branch"]
        branch["revision_url"] = revision_url
        branch["directory_url"] = directory_url

    if origin_info:
        browse_view_name = "browse-origin-branches"
    else:
        browse_view_name = "browse-snapshot-branches"

    prev_branches_url = None
    next_branches_url = None

    if branches_bc:
        query_params_prev = dict(query_params)

        query_params_prev["branches_breadcrumbs"] = ",".join(branches_bc[:-1])
        prev_branches_url = reverse(
            browse_view_name, url_args=url_args, query_params=query_params_prev
        )
    elif branches_from:
        prev_branches_url = reverse(
            browse_view_name, url_args=url_args, query_params=query_params
        )

    if snapshot["next_branch"] is not None:
        query_params_next = dict(query_params)
        next_branch = displayed_branches[-1]["name"]
        del displayed_branches[-1]
        branches_bc.append(next_branch)
        query_params_next["branches_breadcrumbs"] = ",".join(branches_bc)
        next_branches_url = reverse(
            browse_view_name, url_args=url_args, query_params=query_params_next
        )

    heading = "Branches - "
    if origin_info:
        heading += "origin: %s" % origin_info["url"]
    else:
        heading += "snapshot: %s" % snapshot_id

    return render(
        request,
        "browse/branches.html",
        {
            "heading": heading,
            "swh_object_name": "Branches",
            "swh_object_metadata": {},
            "top_right_link": None,
            "displayed_branches": displayed_branches,
            "prev_branches_url": prev_branches_url,
            "next_branches_url": next_branches_url,
            "snapshot_context": snapshot_context,
        },
    )


def browse_snapshot_releases(
    request, snapshot_id=None, origin_url=None, timestamp=None
):
    """
    Django view implementation for browsing a list of releases in a snapshot
    context.
    """
    try:

        _check_origin_url(snapshot_id, origin_url)

        snapshot_context = get_snapshot_context(
            snapshot_id=snapshot_id,
            origin_url=origin_url,
            timestamp=timestamp,
            visit_id=request.GET.get("visit_id"),
        )

        rel_bc = request.GET.get("releases_breadcrumbs", "")
        rel_bc = rel_bc.split(",") if rel_bc else []
        rel_from = rel_bc[-1] if rel_bc else ""

        origin_info = snapshot_context["origin_info"]
        url_args = snapshot_context["url_args"]
        query_params = snapshot_context["query_params"]

        snapshot = service.lookup_snapshot(
            snapshot_context["snapshot_id"],
            rel_from,
            PER_PAGE + 1,
            target_types=["release", "alias"],
        )

        _, displayed_releases = process_snapshot_branches(snapshot)

    except Exception as exc:
        return handle_view_exception(request, exc)

    for release in displayed_releases:
        query_params_tgt = {"snapshot": snapshot_id}
        if origin_info:
            query_params_tgt["origin_url"] = origin_info["url"]

        release_url = reverse(
            "browse-release",
            url_args={"sha1_git": release["id"]},
            query_params=query_params_tgt,
        )

        target_url = ""
        if release["target_type"] == "revision":
            target_url = reverse(
                "browse-revision",
                url_args={"sha1_git": release["target"]},
                query_params=query_params_tgt,
            )
        elif release["target_type"] == "directory":
            target_url = reverse(
                "browse-directory",
                url_args={"sha1_git": release["target"]},
                query_params=query_params_tgt,
            )
        elif release["target_type"] == "content":
            target_url = reverse(
                "browse-content",
                url_args={"query_string": release["target"]},
                query_params=query_params_tgt,
            )
        elif release["target_type"] == "release":
            target_url = reverse(
                "browse-release",
                url_args={"sha1_git": release["target"]},
                query_params=query_params_tgt,
            )

        release["release_url"] = release_url
        release["target_url"] = target_url

    if origin_info:
        browse_view_name = "browse-origin-releases"
    else:
        browse_view_name = "browse-snapshot-releases"

    prev_releases_url = None
    next_releases_url = None

    if rel_bc:
        query_params_prev = dict(query_params)

        query_params_prev["releases_breadcrumbs"] = ",".join(rel_bc[:-1])
        prev_releases_url = reverse(
            browse_view_name, url_args=url_args, query_params=query_params_prev
        )
    elif rel_from:
        prev_releases_url = reverse(
            browse_view_name, url_args=url_args, query_params=query_params
        )

    if snapshot["next_branch"] is not None:
        query_params_next = dict(query_params)
        next_rel = displayed_releases[-1]["branch_name"]
        del displayed_releases[-1]
        rel_bc.append(next_rel)
        query_params_next["releases_breadcrumbs"] = ",".join(rel_bc)
        next_releases_url = reverse(
            browse_view_name, url_args=url_args, query_params=query_params_next
        )

    heading = "Releases - "
    if origin_info:
        heading += "origin: %s" % origin_info["url"]
    else:
        heading += "snapshot: %s" % snapshot_id

    return render(
        request,
        "browse/releases.html",
        {
            "heading": heading,
            "top_panel_visible": False,
            "top_panel_collapsible": False,
            "swh_object_name": "Releases",
            "swh_object_metadata": {},
            "top_right_link": None,
            "displayed_releases": displayed_releases,
            "prev_releases_url": prev_releases_url,
            "next_releases_url": next_releases_url,
            "snapshot_context": snapshot_context,
            "vault_cooking": None,
            "show_actions_menu": False,
        },
    )



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.web.common.utils import reverse


def test_api_1_ping(api_client):
    url = reverse("api-1-ping")

    rv = api_client.get(url)

    assert rv.status_code == 200, rv.data
    assert rv["Content-Type"] == "application/json"
    assert rv.data == "pong"



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import datetime

from django.test import modify_settings
import pytest

from .keycloak_mock import mock_keycloak

from swh.web.common.utils import reverse


@pytest.mark.django_db
@modify_settings(
    MIDDLEWARE={"remove": ["swh.web.auth.middlewares.OIDCSessionRefreshMiddleware"]}
)
def test_oidc_session_refresh_middleware_disabled(client, mocker):
    # authenticate but make session expires immediately
    kc_oidc_mock = mock_keycloak(mocker, exp=int(datetime.now().timestamp()))
    client.login(code="", code_verifier="", redirect_uri="")
    kc_oidc_mock.authorization_code.assert_called()

    url = reverse("swh-web-homepage")
    resp = client.get(url)
    # no redirection for silent refresh
    assert resp.status_code != 302


@pytest.mark.django_db
def test_oidc_session_refresh_middleware_enabled(client, mocker):
    # authenticate but make session expires immediately
    kc_oidc_mock = mock_keycloak(mocker, exp=int(datetime.now().timestamp()))
    client.login(code="", code_verifier="", redirect_uri="")
    kc_oidc_mock.authorization_code.assert_called()

    url = reverse("swh-web-homepage")
    resp = client.get(url)

    # should redirect for silent session refresh
    assert resp.status_code == 302
    silent_refresh_url = reverse(
        "oidc-login", query_params={"next_path": url, "prompt": "none"}
    )
    assert resp["location"] == silent_refresh_url



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.model import hashutil

from swh.web.common import query
from swh.web.common.exc import BadInputExc
import pytest


def test_parse_hash_malformed_query_with_more_than_2_parts():
    with pytest.raises(BadInputExc):
        query.parse_hash("sha1:1234567890987654:other-stuff")


def test_parse_hash_guess_sha1():
    h = "f1d2d2f924e986ac86fdf7b36c94bcdf32beec15"
    r = query.parse_hash(h)
    assert r == ("sha1", hashutil.hash_to_bytes(h))


def test_parse_hash_guess_sha256():
    h = "084C799CD551DD1D8D5C5F9A5D593B2E931F5E36122ee5c793c1d08a19839cc0"
    r = query.parse_hash(h)
    assert r == ("sha256", hashutil.hash_to_bytes(h))


def test_parse_hash_guess_algo_malformed_hash():
    with pytest.raises(BadInputExc):
        query.parse_hash("1234567890987654")


def test_parse_hash_check_sha1():
    h = "f1d2d2f924e986ac86fdf7b36c94bcdf32beec15"
    r = query.parse_hash("sha1:" + h)
    assert r == ("sha1", hashutil.hash_to_bytes(h))


def test_parse_hash_check_sha1_git():
    h = "e1d2d2f924e986ac86fdf7b36c94bcdf32beec15"
    r = query.parse_hash("sha1_git:" + h)
    assert r == ("sha1_git", hashutil.hash_to_bytes(h))


def test_parse_hash_check_sha256():
    h = "084C799CD551DD1D8D5C5F9A5D593B2E931F5E36122ee5c793c1d08a19839cc0"
    r = query.parse_hash("sha256:" + h)
    assert r == ("sha256", hashutil.hash_to_bytes(h))


def test_parse_hash_check_algo_malformed_sha1_hash():
    with pytest.raises(BadInputExc):
        query.parse_hash("sha1:1234567890987654")


def test_parse_hash_check_algo_malformed_sha1_git_hash():
    with pytest.raises(BadInputExc):
        query.parse_hash("sha1_git:1234567890987654")


def test_parse_hash_check_algo_malformed_sha256_hash():
    with pytest.raises(BadInputExc):
        query.parse_hash("sha256:1234567890987654")


def test_parse_hash_check_algo_unknown_one():
    with pytest.raises(BadInputExc):
        query.parse_hash("sha2:1234567890987654")


def test_parse_hash_with_algorithms_or_throws_bad_query(mocker):
    mock_hash = mocker.patch("swh.web.common.query.parse_hash")
    mock_hash.side_effect = BadInputExc("Error input")

    with pytest.raises(BadInputExc) as e:
        query.parse_hash_with_algorithms_or_throws(
            "sha1:blah", ["sha1"], "useless error message for this use case"
        )
    assert e.match("Error input")

    mock_hash.assert_called_once_with("sha1:blah")


def test_parse_hash_with_algorithms_or_throws_bad_algo(mocker):
    mock_hash = mocker.patch("swh.web.common.query.parse_hash")
    mock_hash.return_value = "sha1", "123"

    with pytest.raises(BadInputExc) as e:
        query.parse_hash_with_algorithms_or_throws(
            "sha1:431", ["sha1_git"], "Only sha1_git!"
        )
    assert e.match("Only sha1_git!")

    mock_hash.assert_called_once_with("sha1:431")


def test_parse_hash_with_algorithms(mocker):
    mock_hash = mocker.patch("swh.web.common.query.parse_hash")
    mock_hash.return_value = ("sha256", b"123")

    algo, sha = query.parse_hash_with_algorithms_or_throws(
        "sha256:123", ["sha256", "sha1_git"], "useless error message for this use case"
    )

    assert algo == "sha256"
    assert sha == b"123"

    mock_hash.assert_called_once_with("sha256:123")


def test_parse_uuid4():
    actual_uuid = query.parse_uuid4("7c33636b-8f11-4bda-89d9-ba8b76a42cec")

    assert actual_uuid == "7c33636b-8f11-4bda-89d9-ba8b76a42cec"


def test_parse_uuid4_ko():
    with pytest.raises(BadInputExc) as e:
        query.parse_uuid4("7c33636b-8f11-4bda-89d9-ba8b76a42")
    assert e.match("badly formed hexadecimal UUID string")



# Copyright (C) 2018-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from htmlmin import minify
import sentry_sdk

from swh.web.common.utils import prettify_html


class HtmlPrettifyMiddleware(object):
    """
    Django middleware for prettifying generated HTML in
    development mode.
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)
        if "text/html" in response.get("Content-Type", ""):
            if hasattr(response, "content"):
                content = response.content
                response.content = prettify_html(content)
            elif hasattr(response, "streaming_content"):
                content = b"".join(response.streaming_content)
                response.streaming_content = prettify_html(content)

        return response


class HtmlMinifyMiddleware(object):
    """
    Django middleware for minifying generated HTML in
    production mode.
    """

    def __init__(self, get_response=None):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)
        if "text/html" in response.get("Content-Type", ""):
            try:
                minified_html = minify(
                    response.content.decode("utf-8"), convert_charrefs=False
                )
                response.content = minified_html.encode("utf-8")
            except Exception as exc:
                sentry_sdk.capture_exception(exc)
        return response


class ThrottlingHeadersMiddleware(object):
    """
    Django middleware for inserting rate limiting related
    headers in HTTP response.
    """

    def __init__(self, get_response=None):
        self.get_response = get_response

    def __call__(self, request):
        resp = self.get_response(request)
        if "RateLimit-Limit" in request.META:
            resp["X-RateLimit-Limit"] = request.META["RateLimit-Limit"]
        if "RateLimit-Remaining" in request.META:
            resp["X-RateLimit-Remaining"] = request.META["RateLimit-Remaining"]
        if "RateLimit-Reset" in request.META:
            resp["X-RateLimit-Reset"] = request.META["RateLimit-Reset"]
        return resp



# Copyright (C) 2018-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import random

from hypothesis import given

from swh.model.identifiers import CONTENT, REVISION, SNAPSHOT
from swh.web.common.identifiers import get_swh_persistent_id
from swh.web.common.utils import reverse
from swh.web.tests.django_asserts import assert_contains
from swh.web.tests.strategies import (
    content,
    directory,
    origin,
    revision,
    release,
    snapshot,
)

swh_id_prefix = "swh:1:"


@given(content())
def test_content_id_browse(client, content):
    cnt_sha1_git = content["sha1_git"]
    swh_id = swh_id_prefix + "cnt:" + cnt_sha1_git
    url = reverse("browse-swh-id", url_args={"swh_id": swh_id})

    query_string = "sha1_git:" + cnt_sha1_git
    content_browse_url = reverse(
        "browse-content", url_args={"query_string": query_string}
    )

    resp = client.get(url)

    assert resp.status_code == 302
    assert resp["location"] == content_browse_url


@given(directory())
def test_directory_id_browse(client, directory):
    swh_id = swh_id_prefix + "dir:" + directory
    url = reverse("browse-swh-id", url_args={"swh_id": swh_id})

    directory_browse_url = reverse("browse-directory", url_args={"sha1_git": directory})

    resp = client.get(url)

    assert resp.status_code == 302
    assert resp["location"] == directory_browse_url


@given(revision())
def test_revision_id_browse(client, revision):
    swh_id = swh_id_prefix + "rev:" + revision
    url = reverse("browse-swh-id", url_args={"swh_id": swh_id})

    revision_browse_url = reverse("browse-revision", url_args={"sha1_git": revision})

    resp = client.get(url)

    assert resp.status_code == 302
    assert resp["location"] == revision_browse_url

    query_params = {"origin_url": "https://github.com/user/repo"}

    url = reverse(
        "browse-swh-id", url_args={"swh_id": swh_id}, query_params=query_params
    )

    revision_browse_url = reverse(
        "browse-revision", url_args={"sha1_git": revision}, query_params=query_params
    )

    resp = client.get(url)
    assert resp.status_code == 302
    assert resp["location"] == revision_browse_url


@given(release())
def test_release_id_browse(client, release):
    swh_id = swh_id_prefix + "rel:" + release
    url = reverse("browse-swh-id", url_args={"swh_id": swh_id})

    release_browse_url = reverse("browse-release", url_args={"sha1_git": release})

    resp = client.get(url)

    assert resp.status_code == 302
    assert resp["location"] == release_browse_url

    query_params = {"origin_url": "https://github.com/user/repo"}

    url = reverse(
        "browse-swh-id", url_args={"swh_id": swh_id}, query_params=query_params
    )

    release_browse_url = reverse(
        "browse-release", url_args={"sha1_git": release}, query_params=query_params
    )

    resp = client.get(url)
    assert resp.status_code == 302
    assert resp["location"] == release_browse_url


@given(snapshot())
def test_snapshot_id_browse(client, snapshot):
    swh_id = swh_id_prefix + "snp:" + snapshot
    url = reverse("browse-swh-id", url_args={"swh_id": swh_id})

    snapshot_browse_url = reverse("browse-snapshot", url_args={"snapshot_id": snapshot})

    resp = client.get(url)

    assert resp.status_code == 302
    assert resp["location"] == snapshot_browse_url

    query_params = {"origin_url": "https://github.com/user/repo"}

    url = reverse(
        "browse-swh-id", url_args={"swh_id": swh_id}, query_params=query_params
    )

    release_browse_url = reverse(
        "browse-snapshot", url_args={"snapshot_id": snapshot}, query_params=query_params
    )

    resp = client.get(url)
    assert resp.status_code == 302
    assert resp["location"] == release_browse_url


@given(release())
def test_bad_id_browse(client, release):
    swh_id = swh_id_prefix + "foo:" + release
    url = reverse("browse-swh-id", url_args={"swh_id": swh_id})

    resp = client.get(url)
    assert resp.status_code == 400


@given(content())
def test_content_id_optional_parts_browse(client, content):
    cnt_sha1_git = content["sha1_git"]
    optional_parts = ";lines=4-20;origin=https://github.com/user/repo"
    swh_id = swh_id_prefix + "cnt:" + cnt_sha1_git + optional_parts
    url = reverse("browse-swh-id", url_args={"swh_id": swh_id})

    query_string = "sha1_git:" + cnt_sha1_git
    content_browse_url = reverse(
        "browse-content",
        url_args={"query_string": query_string},
        query_params={"origin_url": "https://github.com/user/repo"},
    )
    content_browse_url += "#L4-L20"

    resp = client.get(url)

    assert resp.status_code == 302
    assert resp["location"] == content_browse_url


@given(release())
def test_origin_id_not_resolvable(client, release):
    swh_id = "swh:1:ori:8068d0075010b590762c6cb5682ed53cb3c13deb"
    url = reverse("browse-swh-id", url_args={"swh_id": swh_id})
    resp = client.get(url)
    assert resp.status_code == 400


@given(origin())
def test_legacy_swhid_browse(archive_data, client, origin):
    snapshot = archive_data.snapshot_get_latest(origin["url"])
    revision = archive_data.snapshot_get_head(snapshot)
    directory = archive_data.revision_get(revision)["directory"]
    directory_content = archive_data.directory_ls(directory)
    directory_file = random.choice(
        [e for e in directory_content if e["type"] == "file"]
    )
    legacy_swhid = get_swh_persistent_id(
        CONTENT,
        directory_file["checksums"]["sha1_git"],
        metadata={"origin": origin["url"]},
    )

    url = reverse("browse-swh-id", url_args={"swh_id": legacy_swhid})
    resp = client.get(url)
    assert resp.status_code == 302

    resp = client.get(resp["location"])

    swhid = get_swh_persistent_id(
        CONTENT,
        directory_file["checksums"]["sha1_git"],
        metadata={
            "origin": origin["url"],
            "visit": get_swh_persistent_id(SNAPSHOT, snapshot),
            "anchor": get_swh_persistent_id(REVISION, revision),
        },
    )

    assert_contains(resp, swhid)



# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json

from django.conf.urls import url
from django.core.paginator import Paginator
from django.http import HttpResponse, HttpResponseForbidden, HttpResponseServerError
from django.shortcuts import render

from rest_framework.decorators import api_view, authentication_classes

from swh.web.api.throttling import throttle_scope
from swh.web.common.exc import ForbiddenExc
from swh.web.common.models import SaveOriginRequest
from swh.web.common.origin_save import (
    create_save_origin_request,
    get_savable_visit_types,
    get_save_origin_requests_from_queryset,
)
from swh.web.common.utils import EnforceCSRFAuthentication


def _origin_save_view(request):
    return render(
        request,
        "misc/origin-save.html",
        {"heading": ("Request the saving of a software origin into " "the archive")},
    )


@api_view(["POST"])
@authentication_classes((EnforceCSRFAuthentication,))
@throttle_scope("swh_save_origin")
def _origin_save_request(request, visit_type, origin_url):
    """
    This view is called through AJAX from the save code now form of swh-web.
    We use DRF here as we want to rate limit the number of submitted requests
    per user to avoid being possibly flooded by bots.
    """
    try:
        response = json.dumps(
            create_save_origin_request(visit_type, origin_url), separators=(",", ": ")
        )
        return HttpResponse(response, content_type="application/json")
    except ForbiddenExc as exc:
        return HttpResponseForbidden(
            json.dumps({"detail": str(exc)}), content_type="application/json"
        )
    except Exception as exc:
        return HttpResponseServerError(
            json.dumps({"detail": str(exc)}), content_type="application/json"
        )


def _visit_save_types_list(request):
    visit_types = json.dumps(get_savable_visit_types(), separators=(",", ": "))
    return HttpResponse(visit_types, content_type="application/json")


def _origin_save_requests_list(request, status):

    if status != "all":
        save_requests = SaveOriginRequest.objects.filter(status=status)
    else:
        save_requests = SaveOriginRequest.objects.all()

    table_data = {}
    table_data["recordsTotal"] = save_requests.count()
    table_data["draw"] = int(request.GET["draw"])

    search_value = request.GET["search[value]"]

    column_order = request.GET["order[0][column]"]
    field_order = request.GET["columns[%s][name]" % column_order]
    order_dir = request.GET["order[0][dir]"]
    if order_dir == "desc":
        field_order = "-" + field_order

    save_requests = save_requests.order_by(field_order)

    length = int(request.GET["length"])
    page = int(request.GET["start"]) / length + 1
    save_requests = get_save_origin_requests_from_queryset(save_requests)
    if search_value:
        save_requests = [
            sr
            for sr in save_requests
            if search_value.lower() in sr["save_request_status"].lower()
            or search_value.lower() in sr["save_task_status"].lower()
            or search_value.lower() in sr["visit_type"].lower()
            or search_value.lower() in sr["origin_url"].lower()
        ]

    table_data["recordsFiltered"] = len(save_requests)
    paginator = Paginator(save_requests, length)
    table_data["data"] = paginator.page(page).object_list
    table_data_json = json.dumps(table_data, separators=(",", ": "))
    return HttpResponse(table_data_json, content_type="application/json")


urlpatterns = [
    url(r"^save/$", _origin_save_view, name="origin-save"),
    url(
        r"^save/(?P<visit_type>.+)/url/(?P<origin_url>.+)/$",
        _origin_save_request,
        name="origin-save-request",
    ),
    url(r"^save/types/list/$", _visit_save_types_list, name="origin-save-types-list"),
    url(
        r"^save/requests/list/(?P<status>.+)/$",
        _origin_save_requests_list,
        name="origin-save-requests-list",
    ),
]



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from itertools import product
import random

from prometheus_client.exposition import CONTENT_TYPE_LATEST

import pytest

from swh.web.common.models import (
    SaveOriginRequest,
    SAVE_REQUEST_ACCEPTED,
    SAVE_REQUEST_REJECTED,
    SAVE_REQUEST_PENDING,
    SAVE_TASK_NOT_YET_SCHEDULED,
    SAVE_TASK_SCHEDULED,
    SAVE_TASK_SUCCEED,
    SAVE_TASK_FAILED,
    SAVE_TASK_RUNNING,
    SAVE_TASK_NOT_CREATED,
)
from swh.web.common.origin_save import (
    get_savable_visit_types,
    ACCEPTED_SAVE_REQUESTS_METRIC,
    SUBMITTED_SAVE_REQUESTS_METRIC,
)
from swh.web.common.utils import reverse
from swh.web.tests.django_asserts import assert_contains


@pytest.mark.django_db
def test_origin_save_metrics(client):
    visit_types = get_savable_visit_types()
    request_statuses = (
        SAVE_REQUEST_ACCEPTED,
        SAVE_REQUEST_REJECTED,
        SAVE_REQUEST_PENDING,
    )

    load_task_statuses = (
        SAVE_TASK_NOT_CREATED,
        SAVE_TASK_NOT_YET_SCHEDULED,
        SAVE_TASK_SCHEDULED,
        SAVE_TASK_SUCCEED,
        SAVE_TASK_FAILED,
        SAVE_TASK_RUNNING,
    )

    for _ in range(random.randint(50, 100)):
        visit_type = random.choice(visit_types)
        request_satus = random.choice(request_statuses)
        load_task_status = random.choice(load_task_statuses)

        SaveOriginRequest.objects.create(
            origin_url="origin",
            visit_type=visit_type,
            status=request_satus,
            loading_task_status=load_task_status,
        )

    url = reverse("metrics-prometheus")
    resp = client.get(url)

    assert resp.status_code == 200
    assert resp["Content-Type"] == CONTENT_TYPE_LATEST

    accepted_requests = SaveOriginRequest.objects.filter(status=SAVE_REQUEST_ACCEPTED)

    labels_set = product(visit_types, load_task_statuses)

    for labels in labels_set:
        sor_count = accepted_requests.filter(
            visit_type=labels[0], loading_task_status=labels[1]
        ).count()

        metric_text = (
            f"{ACCEPTED_SAVE_REQUESTS_METRIC}{{"
            f'load_task_status="{labels[1]}",'
            f'visit_type="{labels[0]}"}} {float(sor_count)}\n'
        )

        assert_contains(resp, metric_text)

    labels_set = product(visit_types, request_statuses)

    for labels in labels_set:
        sor_count = SaveOriginRequest.objects.filter(
            visit_type=labels[0], status=labels[1]
        ).count()

        metric_text = (
            f"{SUBMITTED_SAVE_REQUESTS_METRIC}{{"
            f'status="{labels[1]}",'
            f'visit_type="{labels[0]}"}} {float(sor_count)}\n'
        )

        assert_contains(resp, metric_text)



# Copyright (C) 2018-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import random
import textwrap

from django.utils.html import escape
from hypothesis import given

from swh.web.common.identifiers import get_swh_persistent_id
from swh.web.common.utils import reverse, format_utc_iso_date
from swh.web.tests.django_asserts import assert_contains, assert_template_used
from swh.web.tests.strategies import release, origin_with_releases, unknown_release


@given(release())
def test_release_browse(client, archive_data, release):
    _release_browse_checks(client, release, archive_data)


@given(origin_with_releases())
def test_release_browse_with_origin_snapshot(client, archive_data, origin):
    snapshot = archive_data.snapshot_get_latest(origin["url"])
    release = random.choice(
        [
            b["target"]
            for b in snapshot["branches"].values()
            if b["target_type"] == "release"
        ]
    )

    _release_browse_checks(client, release, archive_data, origin_url=origin["url"])
    _release_browse_checks(client, release, archive_data, snapshot_id=snapshot["id"])
    _release_browse_checks(
        client,
        release,
        archive_data,
        origin_url=origin["url"],
        snapshot_id=snapshot["id"],
    )


@given(unknown_release())
def test_release_browse_not_found(client, archive_data, unknown_release):
    url = reverse("browse-release", url_args={"sha1_git": unknown_release})
    resp = client.get(url)
    assert resp.status_code == 404
    assert_template_used(resp, "error.html")
    err_msg = "Release with sha1_git %s not found" % unknown_release
    assert_contains(resp, err_msg, status_code=404)


@given(release())
def test_release_uppercase(client, release):
    url = reverse(
        "browse-release-uppercase-checksum", url_args={"sha1_git": release.upper()}
    )

    resp = client.get(url)
    assert resp.status_code == 302

    redirect_url = reverse("browse-release", url_args={"sha1_git": release})

    assert resp["location"] == redirect_url


def _release_browse_checks(
    client, release, archive_data, origin_url=None, snapshot_id=None
):
    query_params = {"origin_url": origin_url, "snapshot": snapshot_id}

    url = reverse(
        "browse-release", url_args={"sha1_git": release}, query_params=query_params
    )

    release_data = archive_data.release_get(release)

    resp = client.get(url)

    release_id = release_data["id"]
    release_name = release_data["name"]
    author_name = release_data["author"]["name"]

    release_date = release_data["date"]
    message = release_data["message"]
    target_type = release_data["target_type"]
    target = release_data["target"]

    target_url = reverse(
        "browse-revision", url_args={"sha1_git": target}, query_params=query_params
    )
    message_lines = message.split("\n")

    assert resp.status_code == 200
    assert_template_used(resp, "browse/release.html")
    assert_contains(resp, author_name)
    assert_contains(resp, format_utc_iso_date(release_date))
    assert_contains(
        resp,
        "<h6>%s</h6>%s" % (message_lines[0] or "None", "\n".join(message_lines[1:])),
    )
    assert_contains(resp, release_id)
    assert_contains(resp, release_name)
    assert_contains(resp, target_type)
    assert_contains(resp, '<a href="%s">%s</a>' % (escape(target_url), target))

    swh_rel_id = get_swh_persistent_id("release", release_id)
    swh_rel_id_url = reverse("browse-swh-id", url_args={"swh_id": swh_rel_id})
    assert_contains(resp, swh_rel_id)
    assert_contains(resp, swh_rel_id_url)

    if origin_url:
        browse_origin_url = reverse(
            "browse-origin", query_params={"origin_url": origin_url}
        )
        title = (
            f"Browse archived release for origin\n"
            f'<a href="{browse_origin_url}">\n'
            f"  {origin_url}\n"
            f"</a>"
        )
        indent = " " * 6
    elif snapshot_id:
        swh_snp_id = get_swh_persistent_id("snapshot", snapshot_id)
        swh_snp_id_url = reverse("browse-swh-id", url_args={"swh_id": swh_snp_id})
        title = (
            f"Browse archived release for snapshot\n"
            f'<a href="{swh_snp_id_url}">\n'
            f"  {swh_snp_id}\n"
            f"</a>"
        )
        indent = " " * 6
    else:
        title = (
            f"Browse archived release\n"
            f'<a href="{swh_rel_id_url}">\n'
            f"  {swh_rel_id}\n"
            f"</a>"
        )
        indent = " " * 4

    assert_contains(
        resp, textwrap.indent(title, indent),
    )

    if release_data["target_type"] == "revision":
        if origin_url:
            directory_url = reverse(
                "browse-origin-directory",
                query_params={
                    "origin_url": origin_url,
                    "release": release_data["name"],
                    "snapshot": snapshot_id,
                },
            )
        elif snapshot_id:
            directory_url = reverse(
                "browse-snapshot-directory",
                url_args={"snapshot_id": snapshot_id},
                query_params={"release": release_data["name"],},
            )
        else:
            rev = archive_data.revision_get(release_data["target"])
            directory_url = reverse(
                "browse-directory", url_args={"sha1_git": rev["directory"]}
            )
        assert_contains(resp, escape(directory_url))



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information


realm_public_key = (
    "MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAnqF4xvGjaI54P6WtJvyGayxP8A93u"
    "NcA3TH6jitwmyAalj8dN8/NzK9vrdlSA3Ibvp/XQujPSOP7a35YiYFscEJnogTXQpE/FhZrUY"
    "y21U6ezruVUv4z/ER1cYLb+q5ZI86nXSTNCAbH+lw7rQjlvcJ9KvgHEeA5ALXJ1r55zUmNvuy"
    "5o6ke1G3fXbNSXwF4qlWAzo1o7Ms8qNrNyOG8FPx24dvm9xMH7/08IPvh9KUqlnP8h6olpxHr"
    "drX/q4E+Nzj8Tr8p7Z5CimInls40QuOTIhs6C2SwFHUgQgXl9hB9umiZJlwYEpDv0/LO2zYie"
    "Hl5Lv7Iig4FOIXIVCaDGQIDAQAB"
)

oidc_profile = {
    "access_token": (
        "eyJhbGciOiJSUzI1NiIsInR5cCIgOiAiSldUIiwia2lkIiA6ICJPSnhV"
        "Q0p0TmJQT0NOUGFNNmc3ZU1zY2pqTXhoem9vNGxZaFhsa1c2TWhBIn0."
        "eyJqdGkiOiIzMWZjNTBiNy1iYmU1LTRmNTEtOTFlZi04ZTNlZWM1MTMz"
        "MWUiLCJleHAiOjE1ODI3MjM3MDEsIm5iZiI6MCwiaWF0IjoxNTgyNzIz"
        "MTAxLCJpc3MiOiJodHRwOi8vbG9jYWxob3N0OjgwODAvYXV0aC9yZWFs"
        "bXMvU29mdHdhcmVIZXJpdGFnZSIsImF1ZCI6WyJzd2gtd2ViIiwiYWNj"
        "b3VudCJdLCJzdWIiOiJmZWFjZDM0NC1iNDY4LTRhNjUtYTIzNi0xNGY2"
        "MWU2YjcyMDAiLCJ0eXAiOiJCZWFyZXIiLCJhenAiOiJzd2gtd2ViIiwi"
        "YXV0aF90aW1lIjoxNTgyNzIzMTAwLCJzZXNzaW9uX3N0YXRlIjoiZDgy"
        "YjkwZDEtMGE5NC00ZTc0LWFkNjYtZGQ5NTM0MWM3YjZkIiwiYWNyIjoi"
        "MSIsImFsbG93ZWQtb3JpZ2lucyI6WyIqIl0sInJlYWxtX2FjY2VzcyI6"
        "eyJyb2xlcyI6WyJvZmZsaW5lX2FjY2VzcyIsInVtYV9hdXRob3JpemF0"
        "aW9uIl19LCJyZXNvdXJjZV9hY2Nlc3MiOnsiYWNjb3VudCI6eyJyb2xl"
        "cyI6WyJtYW5hZ2UtYWNjb3VudCIsIm1hbmFnZS1hY2NvdW50LWxpbmtz"
        "Iiwidmlldy1wcm9maWxlIl19fSwic2NvcGUiOiJvcGVuaWQgZW1haWwg"
        "cHJvZmlsZSIsImVtYWlsX3ZlcmlmaWVkIjpmYWxzZSwibmFtZSI6Ikpv"
        "aG4gRG9lIiwiZ3JvdXBzIjpbXSwicHJlZmVycmVkX3VzZXJuYW1lIjoi"
        "am9obmRvZSIsImdpdmVuX25hbWUiOiJKb2huIiwiZmFtaWx5X25hbWUi"
        "OiJEb2UiLCJlbWFpbCI6ImpvaG4uZG9lQGV4YW1wbGUuY29tIn0.neJ-"
        "Pmd87J6Gt0fzDqmXFeoy34Iqb5vNNEEgIKqtqg3moaVkbXrO_9R37DJB"
        "AgdFv0owVONK3GbqPOEICePgG6RFtri999DetNE-O5sB4fwmHPWcHPlO"
        "kcPLbVJqu6zWo-2AzlfAy5bCNvj_wzs2tjFjLeHcRgR1a1WY3uTp5EWc"
        "HITCWQZzZWFGZTZCTlGkpdyJTqxGBdSHRB4NlIVGpYSTBsBsxttFEetl"
        "rpcNd4-5AteFprIr9hn9VasIIF8WdFdtC2e8xGMJW5Q0M3G3Iu-LLNmE"
        "oTIDqtbJ7OrIcGBIwsc3seCV3eCG6kOYwz5w-f8DeOpwcDX58yYPmapJ"
        "6A"
    ),
    "expires_in": 600,
    "id_token": (
        "eyJhbGciOiJSUzI1NiIsInR5cCIgOiAiSldUIiwia2lkIiA6ICJPSnhVQ0p0"
        "TmJQT0NOUGFNNmc3ZU1zY2pqTXhoem9vNGxZaFhsa1c2TWhBIn0.eyJqdGki"
        "OiI0NDRlYzU1My1iYzhiLTQ2YjYtOTlmYS0zOTc3YTJhZDY1ZmEiLCJleHAi"
        "OjE1ODI3MjM3MDEsIm5iZiI6MCwiaWF0IjoxNTgyNzIzMTAxLCJpc3MiOiJo"
        "dHRwOi8vbG9jYWxob3N0OjgwODAvYXV0aC9yZWFsbXMvU29mdHdhcmVIZXJp"
        "dGFnZSIsImF1ZCI6InN3aC13ZWIiLCJzdWIiOiJmZWFjZDM0NC1iNDY4LTRh"
        "NjUtYTIzNi0xNGY2MWU2YjcyMDAiLCJ0eXAiOiJJRCIsImF6cCI6InN3aC13"
        "ZWIiLCJhdXRoX3RpbWUiOjE1ODI3MjMxMDAsInNlc3Npb25fc3RhdGUiOiJk"
        "ODJiOTBkMS0wYTk0LTRlNzQtYWQ2Ni1kZDk1MzQxYzdiNmQiLCJhY3IiOiIx"
        "IiwiZW1haWxfdmVyaWZpZWQiOmZhbHNlLCJuYW1lIjoiSm9obiBEb2UiLCJn"
        "cm91cHMiOltdLCJwcmVmZXJyZWRfdXNlcm5hbWUiOiJqb2huZG9lIiwiZ2l2"
        "ZW5fbmFtZSI6IkpvaG4iLCJmYW1pbHlfbmFtZSI6IkRvZSIsImVtYWlsIjoi"
        "am9obi5kb2VAZXhhbXBsZS5jb20ifQ.YB7bxlz_wgLJSkylVjmqedxQgEMee"
        "JOdi9CFHXV4F3ZWsEZ52CGuJXsozkX2oXvgU06MzzLNEK8ojgrPSNzjRkutL"
        "aaLq_YUzv4iV8fmKUS_aEyiYZbfoBe3Y4dwv2FoPEPCt96iTwpzM5fg_oYw_"
        "PHCq-Yl5SulT1nTrJZpntkf0hRjmxlDO06JMp0aZ8xS8RYJqH48xCRf_DARE"
        "0jJV2-UuzOWI6xBATwFfP44kV6wFmErLN5txMgwZzCSB2OCe5Cl1il0eTQTN"
        "ybeSYZeZE61QtuTRUHeP1D1qSbJGy5g_S67SdTkS-hQFvfrrD84qGflIEqnX"
        "ZbYnitD1Typ6Q"
    ),
    "not-before-policy": 0,
    "refresh_expires_in": 1800,
    "refresh_token": (
        "eyJhbGciOiJIUzI1NiIsInR5cCIgOiAiSldUIiwia2lkIiA6ICJmNjM"
        "zMDE5MS01YTU4LTQxMDAtOGIzYS00ZDdlM2U1NjA3MTgifQ.eyJqdGk"
        "iOiIxYWI5ZWZmMS0xZWZlLTQ3MDMtOGQ2YS03Nzg1NWUwYzQyYTYiLC"
        "JleHAiOjE1ODI3MjQ5MDEsIm5iZiI6MCwiaWF0IjoxNTgyNzIzMTAxL"
        "CJpc3MiOiJodHRwOi8vbG9jYWxob3N0OjgwODAvYXV0aC9yZWFsbXMv"
        "U29mdHdhcmVIZXJpdGFnZSIsImF1ZCI6Imh0dHA6Ly9sb2NhbGhvc3Q"
        "6ODA4MC9hdXRoL3JlYWxtcy9Tb2Z0d2FyZUhlcml0YWdlIiwic3ViIj"
        "oiZmVhY2QzNDQtYjQ2OC00YTY1LWEyMzYtMTRmNjFlNmI3MjAwIiwid"
        "HlwIjoiUmVmcmVzaCIsImF6cCI6InN3aC13ZWIiLCJhdXRoX3RpbWUi"
        "OjAsInNlc3Npb25fc3RhdGUiOiJkODJiOTBkMS0wYTk0LTRlNzQtYWQ"
        "2Ni1kZDk1MzQxYzdiNmQiLCJyZWFsbV9hY2Nlc3MiOnsicm9sZXMiOl"
        "sib2ZmbGluZV9hY2Nlc3MiLCJ1bWFfYXV0aG9yaXphdGlvbiJdfSwic"
        "mVzb3VyY2VfYWNjZXNzIjp7ImFjY291bnQiOnsicm9sZXMiOlsibWFu"
        "YWdlLWFjY291bnQiLCJtYW5hZ2UtYWNjb3VudC1saW5rcyIsInZpZXc"
        "tcHJvZmlsZSJdfX0sInNjb3BlIjoib3BlbmlkIGVtYWlsIHByb2ZpbG"
        "UifQ.xQYrl2CMP_GQ_TFqhsTz-rTs3WuZz5I37toi1eSsDMI"
    ),
    "scope": "openid email profile",
    "session_state": "d82b90d1-0a94-4e74-ad66-dd95341c7b6d",
    "token_type": "bearer",
}

userinfo = {
    "email": "john.doe@example.com",
    "email_verified": False,
    "family_name": "Doe",
    "given_name": "John",
    "groups": ["/staff"],
    "name": "John Doe",
    "preferred_username": "johndoe",
    "sub": "feacd344-b468-4a65-a236-14f61e6b7200",
}



# Copyright (C) 2017-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import functools
from typing import Dict

from pygments.lexers import get_all_lexers, get_lexer_for_filename
import sentry_sdk

# set of languages ids that can be highlighted
# by highlight.js library
_hljs_languages = set(
    [
        "1c",
        "abnf",
        "accesslog",
        "actionscript",
        "ada",
        "angelscript",
        "apache",
        "applescript",
        "arcade",
        "arduino",
        "armasm",
        "asciidoc",
        "aspectj",
        "autohotkey",
        "autoit",
        "avrasm",
        "awk",
        "axapta",
        "bash",
        "basic",
        "bnf",
        "brainfuck",
        "cal",
        "capnproto",
        "ceylon",
        "clean",
        "clojure",
        "clojure-repl",
        "cmake",
        "coffeescript",
        "coq",
        "cos",
        "cpp",
        "crmsh",
        "crystal",
        "cs",
        "csp",
        "css",
        "d",
        "dart",
        "delphi",
        "diff",
        "django",
        "dns",
        "dockerfile",
        "dos",
        "dsconfig",
        "dts",
        "dust",
        "ebnf",
        "elixir",
        "elm",
        "erb",
        "erlang",
        "erlang-repl",
        "excel",
        "fix",
        "flix",
        "fortran",
        "fsharp",
        "gams",
        "gauss",
        "gcode",
        "gherkin",
        "glsl",
        "gml",
        "go",
        "golo",
        "gradle",
        "groovy",
        "haml",
        "handlebars",
        "haskell",
        "haxe",
        "hsp",
        "htmlbars",
        "http",
        "hy",
        "inform7",
        "ini",
        "irpf90",
        "isbl",
        "java",
        "javascript",
        "jboss-cli",
        "json",
        "julia",
        "julia-repl",
        "kotlin",
        "lasso",
        "ldif",
        "leaf",
        "less",
        "lisp",
        "livecodeserver",
        "livescript",
        "llvm",
        "lsl",
        "lua",
        "makefile",
        "markdown",
        "mathematica",
        "matlab",
        "maxima",
        "mel",
        "mercury",
        "mipsasm",
        "mizar",
        "mojolicious",
        "monkey",
        "moonscript",
        "n1ql",
        "nginx",
        "nimrod",
        "nix",
        "nsis",
        "objectivec",
        "ocaml",
        "openscad",
        "oxygene",
        "parser3",
        "perl",
        "pf",
        "pgsql",
        "php",
        "plaintext",
        "pony",
        "powershell",
        "processing",
        "profile",
        "prolog",
        "properties",
        "protobuf",
        "puppet",
        "purebasic",
        "python",
        "q",
        "qml",
        "r",
        "reasonml",
        "rib",
        "roboconf",
        "routeros",
        "rsl",
        "ruby",
        "ruleslanguage",
        "rust",
        "sas",
        "scala",
        "scheme",
        "scilab",
        "scss",
        "shell",
        "smali",
        "smalltalk",
        "sml",
        "sqf",
        "sql",
        "stan",
        "stata",
        "step21",
        "stylus",
        "subunit",
        "swift",
        "taggerscript",
        "tap",
        "tcl",
        "tex",
        "thrift",
        "tp",
        "twig",
        "typescript",
        "vala",
        "vbnet",
        "vbscript",
        "vbscript-html",
        "verilog",
        "vhdl",
        "vim",
        "x86asm",
        "xl",
        "xml",
        "xquery",
        "yaml",
        "zephir",
    ]
)


# languages aliases defined in highlight.js
_hljs_languages_aliases = {
    "ado": "stata",
    "adoc": "asciidoc",
    "ahk": "autohotkey",
    "aj": "aspectj",
    "apacheconf": "apache",
    "arm": "armasm",
    "as": "actionscript",
    "asc": "asciidoc",
    "atom": "xml",
    "bas": "basic",
    "bat": "dos",
    "bf": "brainfuck",
    "bind": "dns",
    "bsl": "1c",
    "c-al": "cal",
    "c": "cpp",
    "c++": "cpp",
    "capnp": "capnproto",
    "cc": "cpp",
    "clj": "clojure",
    "cls": "cos",
    "cmake.in": "cmake",
    "cmd": "dos",
    "coffee": "coffeescript",
    "console": "shell",
    "cr": "crystal",
    "craftcms": "twig",
    "crm": "crmsh",
    "csharp": "cs",
    "cson": "coffeescript",
    "dcl": "clean",
    "dfm": "delphi",
    "do": "stata",
    "docker": "dockerfile",
    "dpr": "delphi",
    "dst": "dust",
    "dtsi": "dts",
    "ep": "mojolicious",
    "erl": "erlang",
    "ex": "elixir",
    "exs": "elixir",
    "f90": "fortran",
    "f95": "fortran",
    "feature": "gherkin",
    "freepascal": "delphi",
    "fs": "fsharp",
    "fsx": "fsharp",
    "gemspec": "ruby",
    "GML": "gml",
    "gms": "gams",
    "golang": "go",
    "graph": "roboconf",
    "gss": "gauss",
    "gyp": "python",
    "h": "cpp",
    "h++": "cpp",
    "hbs": "handlebars",
    "hpp": "cpp",
    "hs": "haskell",
    "html": "xml",
    "html.handlebars": "handlebars",
    "html.hbs": "handlebars",
    "https": "http",
    "hx": "haxe",
    "hylang": "hy",
    "i7": "inform7",
    "i7x": "inform7",
    "iced": "coffeescript",
    "icl": "clean",
    "ino": "arduino",
    "instances": "roboconf",
    "ipynb": "json",
    "irb": "ruby",
    "jinja": "django",
    "js": "javascript",
    "jsp": "java",
    "jsx": "javascript",
    "k": "q",
    "kdb": "q",
    "kt": "kotlin",
    "lassoscript": "lasso",
    "lazarus": "delphi",
    "lc": "livecode",
    "lfm": "delphi",
    "ll": "llvm",
    "lpr": "delphi",
    "ls": "livescript",
    "m": "matlab",
    "mak": "makefile",
    "md": "markdown",
    "mikrotik": "routeros",
    "mips": "mipsasm",
    "mk": "monkey",
    "mkd": "markdown",
    "mkdown": "markdown",
    "ml": "ocaml",
    "mli": "ocaml",
    "mm": "objectivec",
    "mma": "mathematica",
    "moo": "mercury",
    "moon": "moonscript",
    "nav": "cal",
    "nb": "mathematica",
    "nc": "gcode",
    "nginxconf": "nginx",
    "ni": "inform7",
    "nim": "nimrod",
    "nixos": "nix",
    "nsi": "nsis",
    "obj-c": "objectivec",
    "objc": "objectivec",
    "osascript": "applescript",
    "osl": "rsl",
    "p": "parser3",
    "p21": "step21",
    "pas": "delphi",
    "pascal": "delphi",
    "patch": "diff",
    "pb": "purebasic",
    "pbi": "purebasic",
    "pcmk": "crmsh",
    "pde": "processing",
    "pf.conf": "pf",
    "php3": "php",
    "php4": "php",
    "php5": "php",
    "php6": "php",
    "php7": "php",
    "pl": "perl",
    "plist": "xml",
    "pm": "perl",
    "podspec": "ruby",
    "postgres": "pgsql",
    "postgresql": "pgsql",
    "pp": "puppet",
    "proto": "protobuf",
    "ps": "powershell",
    "ps1": "powershell",
    "psd1": "powershell",
    "psm1": "powershell",
    "py": "python",
    "qt": "qml",
    "rb": "ruby",
    "re": "reasonml",
    "rei": "reasonml",
    "rs": "rust",
    "rsc": "routeros",
    "rss": "xml",
    "rst": "nohighlight",
    "s": "armasm",
    "SAS": "sas",
    "scad": "openscad",
    "sci": "scilab",
    "scm": "scheme",
    "sh": "bash",
    "sig": "sml",
    "sl": "rsl",
    "st": "smalltalk",
    "step": "step21",
    "stp": "step21",
    "styl": "stylus",
    "sv": "verilog",
    "svh": "verilog",
    "tao": "xl",
    "thor": "ruby",
    "tk": "tcl",
    "toml": "ini",
    "ts": "typescript",
    "txt": "nohighlight",
    "v": "coq",
    "vb": "vbnet",
    "vbs": "vbscript",
    "vhd": "vhdl",
    "wildfly-cli": "jboss-cli",
    "wl": "mathematica",
    "wls": "mathematica",
    "xhtml": "xml",
    "xjb": "xml",
    "xls": "excel",
    "xlsx": "excel",
    "xpath": "xquery",
    "xpo": "axapta",
    "xpp": "axapta",
    "xq": "xquery",
    "xqy": "xquery",
    "xsd": "xml",
    "xsl": "xml",
    "YAML": "yaml",
    "yml": "yaml",
    "zep": "zephir",
    "zone": "dns",
    "zsh": "bash",
}

# dictionary mapping pygment lexers to hljs languages
_pygments_lexer_to_hljs_language = {}  # type: Dict[str, str]


# dictionary mapping mime types to hljs languages
_mime_type_to_hljs_language = {
    "text/x-c": "cpp",
    "text/x-c++": "cpp",
    "text/x-msdos-batch": "dos",
    "text/x-lisp": "lisp",
    "text/x-shellscript": "bash",
}

# dictionary mapping filenames to hljs languages
_filename_to_hljs_language = {
    "cmakelists.txt": "cmake",
    ".htaccess": "apache",
    "httpd.conf": "apache",
    "access.log": "accesslog",
    "nginx.log": "accesslog",
    "resolv.conf": "dns",
    "dockerfile": "docker",
    "nginx.conf": "nginx",
    "pf.conf": "pf",
}


# function to fill the above dictionaries
def _init_pygments_to_hljs_map():
    if len(_pygments_lexer_to_hljs_language) == 0:
        for lexer in get_all_lexers():
            lexer_name = lexer[0]
            lang_aliases = lexer[1]
            lang_mime_types = lexer[3]
            lang = None
            for lang_alias in lang_aliases:
                if lang_alias in _hljs_languages:
                    lang = lang_alias
                    _pygments_lexer_to_hljs_language[lexer_name] = lang_alias
                    break

            if lang:
                for lang_mime_type in lang_mime_types:
                    _mime_type_to_hljs_language[lang_mime_type] = lang


def get_hljs_language_from_filename(filename):
    """Function that tries to associate a language supported by highlight.js
    from a filename.

    Args:
        filename: input filename

    Returns:
        highlight.js language id or None if no correspondence has been found
    """
    _init_pygments_to_hljs_map()
    if filename:
        filename_lower = filename.lower()
        if filename_lower in _filename_to_hljs_language:
            return _filename_to_hljs_language[filename_lower]
        if filename_lower in _hljs_languages:
            return filename_lower
        exts = filename_lower.split(".")
        # check if file extension matches an hljs language
        # also handle .ext.in cases
        for ext in reversed(exts[-2:]):
            if ext in _hljs_languages:
                return ext
            if ext in _hljs_languages_aliases:
                return _hljs_languages_aliases[ext]

        # otherwise use Pygments language database
        lexer = None
        # try to find a Pygment lexer
        try:
            lexer = get_lexer_for_filename(filename)
        except Exception as exc:
            sentry_sdk.capture_exception(exc)
        # if there is a correspondence between the lexer and an hljs
        # language, return it
        if lexer and lexer.name in _pygments_lexer_to_hljs_language:
            return _pygments_lexer_to_hljs_language[lexer.name]
        # otherwise, try to find a match between the file extensions
        # associated to the lexer and the hljs language aliases
        if lexer:
            exts = [ext.replace("*.", "") for ext in lexer.filenames]
            for ext in exts:
                if ext in _hljs_languages_aliases:
                    return _hljs_languages_aliases[ext]
    return None


def get_hljs_language_from_mime_type(mime_type):
    """Function that tries to associate a language supported by highlight.js
    from a mime type.

    Args:
        mime_type: input mime type

    Returns:
        highlight.js language id or None if no correspondence has been found
    """
    _init_pygments_to_hljs_map()
    if mime_type and mime_type in _mime_type_to_hljs_language:
        return _mime_type_to_hljs_language[mime_type]
    return None


@functools.lru_cache()
def get_supported_languages():
    """
    Return the list of programming languages that can be highlighted using the
    highlight.js library.

    Returns:
        List[str]: the list of supported languages
    """
    return sorted(list(_hljs_languages))



# Copyright (C) 2018-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.web.api.apidoc import api_doc, format_docstring
from swh.web.api.apiurls import api_route
from swh.web.common import service
from swh.web.common.exc import LargePayloadExc
from swh.web.common.identifiers import (
    resolve_swh_persistent_id,
    get_persistent_identifier,
    group_swh_persistent_identifiers,
)


@api_route(r"/resolve/(?P<swh_id>.*)/", "api-1-resolve-swh-pid")
@api_doc("/resolve/")
@format_docstring()
def api_resolve_swh_pid(request, swh_id):
    """
    .. http:get:: /api/1/resolve/(swh_id)/

        Resolve a Software Heritage persistent identifier.

        Try to resolve a provided `persistent identifier
        <https://docs.softwareheritage.org/devel/swh-model/persistent-identifiers.html>`_
        into an url for browsing the pointed archive object. If the provided
        identifier is valid, the existence of the object in the archive
        will also be checked.

        :param string swh_id: a Software Heritage persistent identifier

        :>json string browse_url: the url for browsing the pointed object
        :>json object metadata: object holding optional parts of the
            persistent identifier
        :>json string namespace: the persistent identifier namespace
        :>json string object_id: the hash identifier of the pointed object
        :>json string object_type: the type of the pointed object
        :>json number scheme_version: the scheme version of the persistent
            identifier

        {common_headers}

        :statuscode 200: no error
        :statuscode 400: an invalid persistent identifier has been provided
        :statuscode 404: the pointed object does not exist in the archive

        **Example:**

        .. parsed-literal::

            :swh_web_api:`resolve/swh:1:rev:96db9023b881d7cd9f379b0c154650d6c108e9a3;origin=https://github.com/openssl/openssl/`
    """
    # try to resolve the provided pid
    swh_id_resolved = resolve_swh_persistent_id(swh_id)
    # id is well-formed, now check that the pointed
    # object is present in the archive, NotFoundExc
    # will be raised otherwise
    swh_id_parsed = swh_id_resolved["swh_id_parsed"]
    object_type = swh_id_parsed.object_type
    object_id = swh_id_parsed.object_id
    service.lookup_object(object_type, object_id)
    # id is well-formed and the pointed object exists
    swh_id_data = swh_id_parsed._asdict()
    swh_id_data["browse_url"] = request.build_absolute_uri(
        swh_id_resolved["browse_url"]
    )
    return swh_id_data


@api_route(r"/known/", "api-1-known", methods=["POST"])
@api_doc("/known/")
@format_docstring()
def api_swh_pid_known(request):
    """
    .. http:post:: /api/1/known/

        Check if a list of objects are present in the Software Heritage
        archive.

        The objects to check existence must be provided using Software Heritage
        `persistent identifiers
        <https://docs.softwareheritage.org/devel/swh-model/persistent-identifiers.html>`_.

        :<jsonarr string -: input array of Software Heritage persistent
            identifiers, its length can not exceed 1000.

        :>json object <swh_pid>: an object whose keys are input persistent
            identifiers and values objects with the following keys:

                * **known (bool)**: whether the object was found

        {common_headers}

        :statuscode 200: no error
        :statuscode 400: an invalid persistent identifier was provided
        :statuscode 413: the input array of persistent identifiers is too large

    """
    limit = 1000
    if len(request.data) > limit:
        raise LargePayloadExc(
            "The maximum number of PIDs this endpoint can " "receive is %s" % limit
        )

    persistent_ids = [get_persistent_identifier(pid) for pid in request.data]

    response = {str(pid): {"known": False} for pid in persistent_ids}

    # group pids by their type
    pids_by_type = group_swh_persistent_identifiers(persistent_ids)
    # search for hashes not present in the storage
    missing_hashes = service.lookup_missing_hashes(pids_by_type)

    for pid in persistent_ids:
        if pid.object_id not in missing_hashes:
            response[str(pid)]["known"] = True

    return response



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import datetime, timedelta

from django.contrib.auth import authenticate, get_backends

import pytest

from django.conf import settings

from rest_framework.exceptions import AuthenticationFailed

from swh.web.auth.backends import OIDCBearerTokenAuthentication
from swh.web.auth.models import OIDCUser
from swh.web.common.utils import reverse

from . import sample_data
from .keycloak_mock import mock_keycloak


def _authenticate_user(request_factory):
    request = request_factory.get(reverse("oidc-login-complete"))

    return authenticate(
        request=request,
        code="some-code",
        code_verifier="some-code-verifier",
        redirect_uri="https://localhost:5004",
    )


def _check_authenticated_user(user, decoded_token):
    assert user is not None
    assert isinstance(user, OIDCUser)
    assert user.id != 0
    assert user.username == decoded_token["preferred_username"]
    assert user.password == ""
    assert user.first_name == decoded_token["given_name"]
    assert user.last_name == decoded_token["family_name"]
    assert user.email == decoded_token["email"]
    assert user.is_staff == ("/staff" in decoded_token["groups"])
    assert user.sub == decoded_token["sub"]


@pytest.mark.django_db
def test_oidc_code_pkce_auth_backend_success(mocker, request_factory):
    kc_oidc_mock = mock_keycloak(mocker)
    oidc_profile = sample_data.oidc_profile
    user = _authenticate_user(request_factory)

    decoded_token = kc_oidc_mock.decode_token(user.access_token)
    _check_authenticated_user(user, decoded_token)

    auth_datetime = datetime.fromtimestamp(decoded_token["auth_time"])
    exp_datetime = datetime.fromtimestamp(decoded_token["exp"])
    refresh_exp_datetime = auth_datetime + timedelta(
        seconds=oidc_profile["refresh_expires_in"]
    )

    assert user.access_token == oidc_profile["access_token"]
    assert user.expires_at == exp_datetime
    assert user.id_token == oidc_profile["id_token"]
    assert user.refresh_token == oidc_profile["refresh_token"]
    assert user.refresh_expires_at == refresh_exp_datetime
    assert user.scope == oidc_profile["scope"]
    assert user.session_state == oidc_profile["session_state"]

    backend_path = "swh.web.auth.backends.OIDCAuthorizationCodePKCEBackend"
    assert user.backend == backend_path
    backend_idx = settings.AUTHENTICATION_BACKENDS.index(backend_path)
    assert get_backends()[backend_idx].get_user(user.id) == user


@pytest.mark.django_db
def test_oidc_code_pkce_auth_backend_failure(mocker, request_factory):
    mock_keycloak(mocker, auth_success=False)

    user = _authenticate_user(request_factory)

    assert user is None


@pytest.mark.django_db
def test_drf_oidc_bearer_token_auth_backend_success(mocker, api_request_factory):
    url = reverse("api-1-stat-counters")
    drf_auth_backend = OIDCBearerTokenAuthentication()

    kc_oidc_mock = mock_keycloak(mocker)

    access_token = sample_data.oidc_profile["access_token"]
    decoded_token = kc_oidc_mock.decode_token(access_token)

    request = api_request_factory.get(url, HTTP_AUTHORIZATION=f"Bearer {access_token}")

    user, _ = drf_auth_backend.authenticate(request)
    _check_authenticated_user(user, decoded_token)
    # oidc_profile is not filled when authenticating through bearer token
    assert hasattr(user, "access_token") and user.access_token is None


@pytest.mark.django_db
def test_drf_oidc_bearer_token_auth_backend_failure(mocker, api_request_factory):

    url = reverse("api-1-stat-counters")
    drf_auth_backend = OIDCBearerTokenAuthentication()

    # simulate a failed authentication with a bearer token in expected format
    mock_keycloak(mocker, auth_success=False)

    access_token = sample_data.oidc_profile["access_token"]

    request = api_request_factory.get(url, HTTP_AUTHORIZATION=f"Bearer {access_token}")

    with pytest.raises(AuthenticationFailed):
        drf_auth_backend.authenticate(request)

    # simulate a failed authentication with an invalid bearer token format
    mock_keycloak(mocker)

    request = api_request_factory.get(
        url, HTTP_AUTHORIZATION="Bearer invalid-token-format"
    )

    with pytest.raises(AuthenticationFailed):
        drf_auth_backend.authenticate(request)


def test_drf_oidc_auth_invalid_or_missing_auth_type(api_request_factory):

    url = reverse("api-1-stat-counters")
    drf_auth_backend = OIDCBearerTokenAuthentication()

    access_token = sample_data.oidc_profile["access_token"]

    # Invalid authorization type
    request = api_request_factory.get(url, HTTP_AUTHORIZATION="Foo token")

    with pytest.raises(AuthenticationFailed):
        drf_auth_backend.authenticate(request)

    # Missing authorization type
    request = api_request_factory.get(url, HTTP_AUTHORIZATION=f"{access_token}")

    with pytest.raises(AuthenticationFailed):
        drf_auth_backend.authenticate(request)



# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from hypothesis import given

import pytest

from swh.model.hashutil import hash_to_hex
from swh.web.common.exc import NotFoundExc
from swh.web.common.origin_visits import get_origin_visits, get_origin_visit
from swh.web.tests.strategies import new_snapshots


@given(new_snapshots(3))
def test_get_origin_visits(mocker, snapshots):
    mock_service = mocker.patch("swh.web.common.service")
    mock_service.MAX_LIMIT = 2

    def _lookup_origin_visits(*args, **kwargs):
        if kwargs["last_visit"] is None:
            return [
                {
                    "visit": 1,
                    "date": "2017-05-06T00:59:10+00:00",
                    "status": "full",
                    "snapshot": hash_to_hex(snapshots[0].id),
                    "type": "git",
                },
                {
                    "visit": 2,
                    "date": "2017-08-06T00:59:10+00:00",
                    "status": "full",
                    "snapshot": hash_to_hex(snapshots[1].id),
                    "type": "git",
                },
            ]
        else:
            return [
                {
                    "visit": 3,
                    "date": "2017-09-06T00:59:10+00:00",
                    "status": "full",
                    "snapshot": hash_to_hex(snapshots[2].id),
                    "type": "git",
                }
            ]

    mock_service.lookup_origin_visits.side_effect = _lookup_origin_visits

    origin_info = {
        "url": "https://github.com/foo/bar",
    }

    origin_visits = get_origin_visits(origin_info)

    assert len(origin_visits) == 3


@given(new_snapshots(5))
def test_get_origin_visit(mocker, snapshots):
    mock_origin_visits = mocker.patch("swh.web.common.origin_visits.get_origin_visits")
    origin_info = {
        "url": "https://github.com/foo/bar",
    }
    visits = [
        {
            "status": "full",
            "date": "2015-07-09T21:09:24+00:00",
            "visit": 1,
            "origin": "https://github.com/foo/bar",
            "type": "git",
            "snapshot": hash_to_hex(snapshots[0].id),
        },
        {
            "status": "full",
            "date": "2016-02-23T18:05:23.312045+00:00",
            "visit": 2,
            "origin": "https://github.com/foo/bar",
            "type": "git",
            "snapshot": hash_to_hex(snapshots[1].id),
        },
        {
            "status": "full",
            "date": "2016-03-28T01:35:06.554111+00:00",
            "visit": 3,
            "origin": "https://github.com/foo/bar",
            "type": "git",
            "snapshot": hash_to_hex(snapshots[2].id),
        },
        {
            "status": "full",
            "date": "2016-06-18T01:22:24.808485+00:00",
            "visit": 4,
            "origin": "https://github.com/foo/bar",
            "type": "git",
            "snapshot": hash_to_hex(snapshots[3].id),
        },
        {
            "status": "full",
            "date": "2016-08-14T12:10:00.536702+00:00",
            "visit": 5,
            "origin": "https://github.com/foo/bar",
            "type": "git",
            "snapshot": hash_to_hex(snapshots[4].id),
        },
    ]
    mock_origin_visits.return_value = visits

    visit_id = 12
    with pytest.raises(NotFoundExc) as e:
        visit = get_origin_visit(origin_info, visit_id=visit_id)

    assert e.match("Visit with id %s" % visit_id)
    assert e.match("url %s" % origin_info["url"])

    visit = get_origin_visit(origin_info, visit_id=2)
    assert visit == visits[1]

    visit = get_origin_visit(origin_info, visit_ts="2016-02-23T18:05:23.312045+00:00")
    assert visit == visits[1]

    visit = get_origin_visit(origin_info, visit_ts="2016-02-20")
    assert visit == visits[1]

    visit = get_origin_visit(origin_info, visit_ts="2016-06-18T01:22")
    assert visit == visits[3]

    visit = get_origin_visit(origin_info, visit_ts="2016-06-18 01:22")
    assert visit == visits[3]

    visit = get_origin_visit(origin_info, visit_ts=1466208000)
    assert visit == visits[3]

    visit = get_origin_visit(origin_info, visit_ts="2014-01-01")
    assert visit == visits[0]

    visit = get_origin_visit(origin_info, visit_ts="2018-01-01")
    assert visit == visits[-1]


def test_get_origin_visit_latest_valid_snapshot(mocker):
    mock_origin_visits = mocker.patch("swh.web.common.origin_visits.get_origin_visits")
    origin_info = {
        "url": "https://nix-community.github.io/nixpkgs-swh/sources-unstable.json",
    }
    visits = [
        {
            "date": "2020-04-15T12:42:52.936520+00:00",
            "origin": origin_info["url"],
            "snapshot": "d820451681c74eec63693b6ea4e4b8417c76bb7a",
            "status": "partial",
            "type": "nixguix",
            "visit": 16,
        },
        {
            "date": "2020-04-17T17:25:13.738789+00:00",
            "origin": origin_info["url"],
            "snapshot": "d20627c1ae2b5e553e8adcf625f37e37cc5190dd",
            "status": "partial",
            "type": "nixguix",
            "visit": 17,
        },
        {
            "date": "2020-04-19T19:02:42.906079+00:00",
            "origin": origin_info["url"],
            "snapshot": None,
            "status": "partial",
            "type": "nixguix",
            "visit": 18,
        },
        {
            "date": "2020-04-20T12:43:41.120422+00:00",
            "origin": origin_info["url"],
            "snapshot": None,
            "status": "partial",
            "type": "nixguix",
            "visit": 19,
        },
        {
            "date": "2020-04-20T12:46:40.255418+00:00",
            "origin": origin_info["url"],
            "snapshot": None,
            "status": "partial",
            "type": "nixguix",
            "visit": 20,
        },
    ]

    mock_origin_visits.return_value = visits

    visit = get_origin_visit(origin_info)

    assert visit["snapshot"] is not None
    assert visit["visit"] == 17
